# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: bitfount/federated/transport/protos/messages.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor.FileDescriptor(
    name="bitfount/federated/transport/protos/messages.proto",
    package="messages",
    syntax="proto3",
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
    serialized_pb=b'\n2bitfount/federated/transport/protos/messages.proto\x12\x08messages"\x9c\x01\n\x07PodData\x12\x0f\n\x07podName\x18\x01 \x01(\t\x12\x11\n\tprocessor\x18\x02 \x01(\t\x12\r\n\x05podOS\x18\x03 \x01(\t\x12\x10\n\x08\x63puCount\x18\x04 \x01(\r\x12\x18\n\x10totalMemoryBytes\x18\x05 \x01(\x04\x12\x10\n\x08gpuCount\x18\x06 \x01(\r\x12\x14\n\x07gpuName\x18\x07 \x01(\tH\x00\x88\x01\x01\x42\n\n\x08_gpuName"u\n\x0c\x42itfountTask\x12\x15\n\rpodIdentifier\x18\x01 \x01(\t\x12\x1a\n\rencryptedTask\x18\x03 \x01(\x0cH\x00\x88\x01\x01\x12\x14\n\x07taskURL\x18\x04 \x01(\x0cH\x01\x88\x01\x01\x42\x10\n\x0e_encryptedTaskB\n\n\x08_taskURL"6\n\rBitfountTasks\x12%\n\x05tasks\x18\x01 \x03(\x0b\x32\x16.messages.BitfountTask"\xb9\x01\n\x14\x43ommunicationDetails\x12\x11\n\tmailboxId\x18\x01 \x01(\t\x12H\n\rpodMailboxIds\x18\x02 \x03(\x0b\x32\x31.messages.CommunicationDetails.PodMailboxIdsEntry\x12\x0e\n\x06taskId\x18\x03 \x01(\t\x1a\x34\n\x12PodMailboxIdsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01"\xbb\x07\n\x0f\x42itfountMessage\x12:\n\x0bmessageType\x18\x01 \x01(\x0e\x32%.messages.BitfountMessage.MessageType\x12\x0c\n\x04\x62ody\x18\x02 \x01(\x0c\x12\x11\n\trecipient\x18\x03 \x01(\t\x12\x1a\n\x12recipientMailboxId\x18\x04 \x01(\t\x12\x13\n\x06sender\x18\x05 \x01(\tH\x00\x88\x01\x01\x12\x17\n\x0fsenderMailboxId\x18\x06 \x01(\t\x12\x11\n\ttimestamp\x18\x07 \x01(\t\x12\x43\n\rpodMailboxIds\x18\x08 \x03(\x0b\x32,.messages.BitfountMessage.PodMailboxIdsEntry\x12\x1a\n\rreceiptHandle\x18\t \x01(\tH\x01\x88\x01\x01\x12\x13\n\x06taskId\x18\n \x01(\tH\x02\x88\x01\x01\x1a\x34\n\x12PodMailboxIdsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01"\x99\x04\n\x0bMessageType\x12\r\n\tUNDEFINED\x10\x00\x12\x16\n\x12\x41LGORITHM_EXCHANGE\x10\x01\x12\x0f\n\x0bJOB_REQUEST\x10\x02\x12\x0e\n\nJOB_ACCEPT\x10\x03\x12\x0e\n\nJOB_REJECT\x10\x04\x12\x13\n\x0fTRAINING_UPDATE\x10\x05\x12\x15\n\x11TRAINING_COMPLETE\x10\x06\x12\x16\n\x12\x45VALUATION_RESULTS\x10\x07\x12\x10\n\x0cSECURE_SHARE\x10\x08\x12\x14\n\x10MODEL_PARAMETERS\x10\t\x12\x0e\n\nTASK_ABORT\x10\n\x12 \n\x1cLARGE_OBJECT_STORAGE_REQUEST\x10\x0b\x12\x10\n\x0cKEY_EXCHANGE\x10\x0c\x12\x0c\n\x08POD_DATA\x10\r\x12\x10\n\x0cSAML_REQUEST\x10\x0e\x12\x11\n\rSAML_RESPONSE\x10\x0f\x12\x0f\n\x0bLOG_MESSAGE\x10\x10\x12\x14\n\x10TRAINING_METRICS\x10\x11\x12\x11\n\rTASK_COMPLETE\x10\x12\x12\x12\n\x0eOIDC_CHALLENGE\x10\x13\x12\x1a\n\x16OIDC_AFC_PKCE_RESPONSE\x10\x14\x12\x1d\n\x19OIDC_DEVICE_CODE_RESPONSE\x10\x15\x12\x10\n\x0cONLINE_CHECK\x10\x16\x12\x13\n\x0fONLINE_RESPONSE\x10\x17\x12\x0f\n\x0bPSI_DATASET\x10\x18\x12\x0e\n\nTASK_START\x10\x19\x42\t\n\x07_senderB\x10\n\x0e_receiptHandleB\t\n\x07_taskId"i\n\x0f\x41\x63knowledgement\x12\x11\n\tmailboxId\x18\x01 \x01(\t\x12\x15\n\rreceiptHandle\x18\x03 \x01(\t\x12\x1a\n\rdeleteMailbox\x18\x04 \x01(\x08H\x00\x88\x01\x01\x42\x10\n\x0e_deleteMailbox"\x11\n\x0fSuccessResponse"\x85\x01\n\x13LargeStorageRequest\x12\x17\n\x0fsenderMailboxId\x18\x01 \x01(\t\x12\x14\n\x07podName\x18\x02 \x01(\tH\x00\x88\x01\x01\x12\x13\n\x0b\x63ontentSize\x18\x03 \x01(\r\x12\x13\n\x06taskId\x18\x04 \x01(\tH\x01\x88\x01\x01\x42\n\n\x08_podNameB\t\n\x07_taskId"_\n\x17LargeStorageTaskRequest\x12\x0f\n\x07podName\x18\x01 \x01(\t\x12\x13\n\x0b\x63ontentSize\x18\x02 \x01(\r\x12\x13\n\x06taskId\x18\x03 \x01(\tH\x00\x88\x01\x01\x42\t\n\x07_taskId"\xb1\x01\n\x0f\x42lobStorageData\x12\x11\n\tuploadUrl\x18\x01 \x01(\t\x12\x13\n\x0b\x64ownloadUrl\x18\x02 \x01(\t\x12\x41\n\x0cuploadFields\x18\x03 \x03(\x0b\x32+.messages.BlobStorageData.UploadFieldsEntry\x1a\x33\n\x11UploadFieldsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x32\xe4\x03\n\x0eMessageService\x12<\n\nPodConnect\x12\x11.messages.PodData\x1a\x19.messages.SuccessResponse"\x00\x12O\n\x12SetupTaskMailboxes\x12\x17.messages.BitfountTasks\x1a\x1e.messages.CommunicationDetails"\x00\x12L\n\x12\x41\x63knowledgeMessage\x12\x19.messages.Acknowledgement\x1a\x19.messages.SuccessResponse"\x00\x12Q\n\x12GetBitfountMessage\x12\x1e.messages.CommunicationDetails\x1a\x19.messages.BitfountMessage"\x00\x12M\n\x13SendBitfountMessage\x12\x19.messages.BitfountMessage\x1a\x19.messages.SuccessResponse"\x00\x12S\n\x15GetLargeObjectStorage\x12\x1d.messages.LargeStorageRequest\x1a\x19.messages.BlobStorageData"\x00\x62\x06proto3',
)


_BITFOUNTMESSAGE_MESSAGETYPE = _descriptor.EnumDescriptor(
    name="MessageType",
    full_name="messages.BitfountMessage.MessageType",
    filename=None,
    file=DESCRIPTOR,
    create_key=_descriptor._internal_create_key,
    values=[
        _descriptor.EnumValueDescriptor(
            name="UNDEFINED",
            index=0,
            number=0,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="ALGORITHM_EXCHANGE",
            index=1,
            number=1,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="JOB_REQUEST",
            index=2,
            number=2,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="JOB_ACCEPT",
            index=3,
            number=3,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="JOB_REJECT",
            index=4,
            number=4,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="TRAINING_UPDATE",
            index=5,
            number=5,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="TRAINING_COMPLETE",
            index=6,
            number=6,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="EVALUATION_RESULTS",
            index=7,
            number=7,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="SECURE_SHARE",
            index=8,
            number=8,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="MODEL_PARAMETERS",
            index=9,
            number=9,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="TASK_ABORT",
            index=10,
            number=10,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="LARGE_OBJECT_STORAGE_REQUEST",
            index=11,
            number=11,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="KEY_EXCHANGE",
            index=12,
            number=12,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="POD_DATA",
            index=13,
            number=13,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="SAML_REQUEST",
            index=14,
            number=14,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="SAML_RESPONSE",
            index=15,
            number=15,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="LOG_MESSAGE",
            index=16,
            number=16,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="TRAINING_METRICS",
            index=17,
            number=17,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="TASK_COMPLETE",
            index=18,
            number=18,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="OIDC_CHALLENGE",
            index=19,
            number=19,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="OIDC_AFC_PKCE_RESPONSE",
            index=20,
            number=20,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="OIDC_DEVICE_CODE_RESPONSE",
            index=21,
            number=21,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="ONLINE_CHECK",
            index=22,
            number=22,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="ONLINE_RESPONSE",
            index=23,
            number=23,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="PSI_DATASET",
            index=24,
            number=24,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="TASK_START",
            index=25,
            number=25,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    containing_type=None,
    serialized_options=None,
    serialized_start=965,
    serialized_end=1502,
)
_sym_db.RegisterEnumDescriptor(_BITFOUNTMESSAGE_MESSAGETYPE)


_PODDATA = _descriptor.Descriptor(
    name="PodData",
    full_name="messages.PodData",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="podName",
            full_name="messages.PodData.podName",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="processor",
            full_name="messages.PodData.processor",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="podOS",
            full_name="messages.PodData.podOS",
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="cpuCount",
            full_name="messages.PodData.cpuCount",
            index=3,
            number=4,
            type=13,
            cpp_type=3,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="totalMemoryBytes",
            full_name="messages.PodData.totalMemoryBytes",
            index=4,
            number=5,
            type=4,
            cpp_type=4,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="gpuCount",
            full_name="messages.PodData.gpuCount",
            index=5,
            number=6,
            type=13,
            cpp_type=3,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="gpuName",
            full_name="messages.PodData.gpuName",
            index=6,
            number=7,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[
        _descriptor.OneofDescriptor(
            name="_gpuName",
            full_name="messages.PodData._gpuName",
            index=0,
            containing_type=None,
            create_key=_descriptor._internal_create_key,
            fields=[],
        ),
    ],
    serialized_start=65,
    serialized_end=221,
)


_BITFOUNTTASK = _descriptor.Descriptor(
    name="BitfountTask",
    full_name="messages.BitfountTask",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="podIdentifier",
            full_name="messages.BitfountTask.podIdentifier",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="encryptedTask",
            full_name="messages.BitfountTask.encryptedTask",
            index=1,
            number=3,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"",
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="taskURL",
            full_name="messages.BitfountTask.taskURL",
            index=2,
            number=4,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"",
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[
        _descriptor.OneofDescriptor(
            name="_encryptedTask",
            full_name="messages.BitfountTask._encryptedTask",
            index=0,
            containing_type=None,
            create_key=_descriptor._internal_create_key,
            fields=[],
        ),
        _descriptor.OneofDescriptor(
            name="_taskURL",
            full_name="messages.BitfountTask._taskURL",
            index=1,
            containing_type=None,
            create_key=_descriptor._internal_create_key,
            fields=[],
        ),
    ],
    serialized_start=223,
    serialized_end=340,
)


_BITFOUNTTASKS = _descriptor.Descriptor(
    name="BitfountTasks",
    full_name="messages.BitfountTasks",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="tasks",
            full_name="messages.BitfountTasks.tasks",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=342,
    serialized_end=396,
)


_COMMUNICATIONDETAILS_PODMAILBOXIDSENTRY = _descriptor.Descriptor(
    name="PodMailboxIdsEntry",
    full_name="messages.CommunicationDetails.PodMailboxIdsEntry",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="key",
            full_name="messages.CommunicationDetails.PodMailboxIdsEntry.key",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="value",
            full_name="messages.CommunicationDetails.PodMailboxIdsEntry.value",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=b"8\001",
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=532,
    serialized_end=584,
)

_COMMUNICATIONDETAILS = _descriptor.Descriptor(
    name="CommunicationDetails",
    full_name="messages.CommunicationDetails",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="mailboxId",
            full_name="messages.CommunicationDetails.mailboxId",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="podMailboxIds",
            full_name="messages.CommunicationDetails.podMailboxIds",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="taskId",
            full_name="messages.CommunicationDetails.taskId",
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[
        _COMMUNICATIONDETAILS_PODMAILBOXIDSENTRY,
    ],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=399,
    serialized_end=584,
)


_BITFOUNTMESSAGE_PODMAILBOXIDSENTRY = _descriptor.Descriptor(
    name="PodMailboxIdsEntry",
    full_name="messages.BitfountMessage.PodMailboxIdsEntry",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="key",
            full_name="messages.BitfountMessage.PodMailboxIdsEntry.key",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="value",
            full_name="messages.BitfountMessage.PodMailboxIdsEntry.value",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=b"8\001",
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=532,
    serialized_end=584,
)

_BITFOUNTMESSAGE = _descriptor.Descriptor(
    name="BitfountMessage",
    full_name="messages.BitfountMessage",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="messageType",
            full_name="messages.BitfountMessage.messageType",
            index=0,
            number=1,
            type=14,
            cpp_type=8,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="body",
            full_name="messages.BitfountMessage.body",
            index=1,
            number=2,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"",
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="recipient",
            full_name="messages.BitfountMessage.recipient",
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="recipientMailboxId",
            full_name="messages.BitfountMessage.recipientMailboxId",
            index=3,
            number=4,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="sender",
            full_name="messages.BitfountMessage.sender",
            index=4,
            number=5,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="senderMailboxId",
            full_name="messages.BitfountMessage.senderMailboxId",
            index=5,
            number=6,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="timestamp",
            full_name="messages.BitfountMessage.timestamp",
            index=6,
            number=7,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="podMailboxIds",
            full_name="messages.BitfountMessage.podMailboxIds",
            index=7,
            number=8,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="receiptHandle",
            full_name="messages.BitfountMessage.receiptHandle",
            index=8,
            number=9,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="taskId",
            full_name="messages.BitfountMessage.taskId",
            index=9,
            number=10,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[
        _BITFOUNTMESSAGE_PODMAILBOXIDSENTRY,
    ],
    enum_types=[
        _BITFOUNTMESSAGE_MESSAGETYPE,
    ],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[
        _descriptor.OneofDescriptor(
            name="_sender",
            full_name="messages.BitfountMessage._sender",
            index=0,
            containing_type=None,
            create_key=_descriptor._internal_create_key,
            fields=[],
        ),
        _descriptor.OneofDescriptor(
            name="_receiptHandle",
            full_name="messages.BitfountMessage._receiptHandle",
            index=1,
            containing_type=None,
            create_key=_descriptor._internal_create_key,
            fields=[],
        ),
        _descriptor.OneofDescriptor(
            name="_taskId",
            full_name="messages.BitfountMessage._taskId",
            index=2,
            containing_type=None,
            create_key=_descriptor._internal_create_key,
            fields=[],
        ),
    ],
    serialized_start=587,
    serialized_end=1542,
)


_ACKNOWLEDGEMENT = _descriptor.Descriptor(
    name="Acknowledgement",
    full_name="messages.Acknowledgement",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="mailboxId",
            full_name="messages.Acknowledgement.mailboxId",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="receiptHandle",
            full_name="messages.Acknowledgement.receiptHandle",
            index=1,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="deleteMailbox",
            full_name="messages.Acknowledgement.deleteMailbox",
            index=2,
            number=4,
            type=8,
            cpp_type=7,
            label=1,
            has_default_value=False,
            default_value=False,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[
        _descriptor.OneofDescriptor(
            name="_deleteMailbox",
            full_name="messages.Acknowledgement._deleteMailbox",
            index=0,
            containing_type=None,
            create_key=_descriptor._internal_create_key,
            fields=[],
        ),
    ],
    serialized_start=1544,
    serialized_end=1649,
)


_SUCCESSRESPONSE = _descriptor.Descriptor(
    name="SuccessResponse",
    full_name="messages.SuccessResponse",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1651,
    serialized_end=1668,
)


_LARGESTORAGEREQUEST = _descriptor.Descriptor(
    name="LargeStorageRequest",
    full_name="messages.LargeStorageRequest",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="senderMailboxId",
            full_name="messages.LargeStorageRequest.senderMailboxId",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="podName",
            full_name="messages.LargeStorageRequest.podName",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="contentSize",
            full_name="messages.LargeStorageRequest.contentSize",
            index=2,
            number=3,
            type=13,
            cpp_type=3,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="taskId",
            full_name="messages.LargeStorageRequest.taskId",
            index=3,
            number=4,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[
        _descriptor.OneofDescriptor(
            name="_podName",
            full_name="messages.LargeStorageRequest._podName",
            index=0,
            containing_type=None,
            create_key=_descriptor._internal_create_key,
            fields=[],
        ),
        _descriptor.OneofDescriptor(
            name="_taskId",
            full_name="messages.LargeStorageRequest._taskId",
            index=1,
            containing_type=None,
            create_key=_descriptor._internal_create_key,
            fields=[],
        ),
    ],
    serialized_start=1671,
    serialized_end=1804,
)


_LARGESTORAGETASKREQUEST = _descriptor.Descriptor(
    name="LargeStorageTaskRequest",
    full_name="messages.LargeStorageTaskRequest",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="podName",
            full_name="messages.LargeStorageTaskRequest.podName",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="contentSize",
            full_name="messages.LargeStorageTaskRequest.contentSize",
            index=1,
            number=2,
            type=13,
            cpp_type=3,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="taskId",
            full_name="messages.LargeStorageTaskRequest.taskId",
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[
        _descriptor.OneofDescriptor(
            name="_taskId",
            full_name="messages.LargeStorageTaskRequest._taskId",
            index=0,
            containing_type=None,
            create_key=_descriptor._internal_create_key,
            fields=[],
        ),
    ],
    serialized_start=1806,
    serialized_end=1901,
)


_BLOBSTORAGEDATA_UPLOADFIELDSENTRY = _descriptor.Descriptor(
    name="UploadFieldsEntry",
    full_name="messages.BlobStorageData.UploadFieldsEntry",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="key",
            full_name="messages.BlobStorageData.UploadFieldsEntry.key",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="value",
            full_name="messages.BlobStorageData.UploadFieldsEntry.value",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=b"8\001",
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=2030,
    serialized_end=2081,
)

_BLOBSTORAGEDATA = _descriptor.Descriptor(
    name="BlobStorageData",
    full_name="messages.BlobStorageData",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="uploadUrl",
            full_name="messages.BlobStorageData.uploadUrl",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="downloadUrl",
            full_name="messages.BlobStorageData.downloadUrl",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="uploadFields",
            full_name="messages.BlobStorageData.uploadFields",
            index=2,
            number=3,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[
        _BLOBSTORAGEDATA_UPLOADFIELDSENTRY,
    ],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1904,
    serialized_end=2081,
)

_PODDATA.oneofs_by_name["_gpuName"].fields.append(_PODDATA.fields_by_name["gpuName"])
_PODDATA.fields_by_name["gpuName"].containing_oneof = _PODDATA.oneofs_by_name[
    "_gpuName"
]
_BITFOUNTTASK.oneofs_by_name["_encryptedTask"].fields.append(
    _BITFOUNTTASK.fields_by_name["encryptedTask"]
)
_BITFOUNTTASK.fields_by_name[
    "encryptedTask"
].containing_oneof = _BITFOUNTTASK.oneofs_by_name["_encryptedTask"]
_BITFOUNTTASK.oneofs_by_name["_taskURL"].fields.append(
    _BITFOUNTTASK.fields_by_name["taskURL"]
)
_BITFOUNTTASK.fields_by_name["taskURL"].containing_oneof = _BITFOUNTTASK.oneofs_by_name[
    "_taskURL"
]
_BITFOUNTTASKS.fields_by_name["tasks"].message_type = _BITFOUNTTASK
_COMMUNICATIONDETAILS_PODMAILBOXIDSENTRY.containing_type = _COMMUNICATIONDETAILS
_COMMUNICATIONDETAILS.fields_by_name[
    "podMailboxIds"
].message_type = _COMMUNICATIONDETAILS_PODMAILBOXIDSENTRY
_BITFOUNTMESSAGE_PODMAILBOXIDSENTRY.containing_type = _BITFOUNTMESSAGE
_BITFOUNTMESSAGE.fields_by_name["messageType"].enum_type = _BITFOUNTMESSAGE_MESSAGETYPE
_BITFOUNTMESSAGE.fields_by_name[
    "podMailboxIds"
].message_type = _BITFOUNTMESSAGE_PODMAILBOXIDSENTRY
_BITFOUNTMESSAGE_MESSAGETYPE.containing_type = _BITFOUNTMESSAGE
_BITFOUNTMESSAGE.oneofs_by_name["_sender"].fields.append(
    _BITFOUNTMESSAGE.fields_by_name["sender"]
)
_BITFOUNTMESSAGE.fields_by_name[
    "sender"
].containing_oneof = _BITFOUNTMESSAGE.oneofs_by_name["_sender"]
_BITFOUNTMESSAGE.oneofs_by_name["_receiptHandle"].fields.append(
    _BITFOUNTMESSAGE.fields_by_name["receiptHandle"]
)
_BITFOUNTMESSAGE.fields_by_name[
    "receiptHandle"
].containing_oneof = _BITFOUNTMESSAGE.oneofs_by_name["_receiptHandle"]
_BITFOUNTMESSAGE.oneofs_by_name["_taskId"].fields.append(
    _BITFOUNTMESSAGE.fields_by_name["taskId"]
)
_BITFOUNTMESSAGE.fields_by_name[
    "taskId"
].containing_oneof = _BITFOUNTMESSAGE.oneofs_by_name["_taskId"]
_ACKNOWLEDGEMENT.oneofs_by_name["_deleteMailbox"].fields.append(
    _ACKNOWLEDGEMENT.fields_by_name["deleteMailbox"]
)
_ACKNOWLEDGEMENT.fields_by_name[
    "deleteMailbox"
].containing_oneof = _ACKNOWLEDGEMENT.oneofs_by_name["_deleteMailbox"]
_LARGESTORAGEREQUEST.oneofs_by_name["_podName"].fields.append(
    _LARGESTORAGEREQUEST.fields_by_name["podName"]
)
_LARGESTORAGEREQUEST.fields_by_name[
    "podName"
].containing_oneof = _LARGESTORAGEREQUEST.oneofs_by_name["_podName"]
_LARGESTORAGEREQUEST.oneofs_by_name["_taskId"].fields.append(
    _LARGESTORAGEREQUEST.fields_by_name["taskId"]
)
_LARGESTORAGEREQUEST.fields_by_name[
    "taskId"
].containing_oneof = _LARGESTORAGEREQUEST.oneofs_by_name["_taskId"]
_LARGESTORAGETASKREQUEST.oneofs_by_name["_taskId"].fields.append(
    _LARGESTORAGETASKREQUEST.fields_by_name["taskId"]
)
_LARGESTORAGETASKREQUEST.fields_by_name[
    "taskId"
].containing_oneof = _LARGESTORAGETASKREQUEST.oneofs_by_name["_taskId"]
_BLOBSTORAGEDATA_UPLOADFIELDSENTRY.containing_type = _BLOBSTORAGEDATA
_BLOBSTORAGEDATA.fields_by_name[
    "uploadFields"
].message_type = _BLOBSTORAGEDATA_UPLOADFIELDSENTRY
DESCRIPTOR.message_types_by_name["PodData"] = _PODDATA
DESCRIPTOR.message_types_by_name["BitfountTask"] = _BITFOUNTTASK
DESCRIPTOR.message_types_by_name["BitfountTasks"] = _BITFOUNTTASKS
DESCRIPTOR.message_types_by_name["CommunicationDetails"] = _COMMUNICATIONDETAILS
DESCRIPTOR.message_types_by_name["BitfountMessage"] = _BITFOUNTMESSAGE
DESCRIPTOR.message_types_by_name["Acknowledgement"] = _ACKNOWLEDGEMENT
DESCRIPTOR.message_types_by_name["SuccessResponse"] = _SUCCESSRESPONSE
DESCRIPTOR.message_types_by_name["LargeStorageRequest"] = _LARGESTORAGEREQUEST
DESCRIPTOR.message_types_by_name["LargeStorageTaskRequest"] = _LARGESTORAGETASKREQUEST
DESCRIPTOR.message_types_by_name["BlobStorageData"] = _BLOBSTORAGEDATA
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

PodData = _reflection.GeneratedProtocolMessageType(
    "PodData",
    (_message.Message,),
    {
        "DESCRIPTOR": _PODDATA,
        "__module__": "bitfount.federated.transport.protos.messages_pb2"
        # @@protoc_insertion_point(class_scope:messages.PodData)
    },
)
_sym_db.RegisterMessage(PodData)

BitfountTask = _reflection.GeneratedProtocolMessageType(
    "BitfountTask",
    (_message.Message,),
    {
        "DESCRIPTOR": _BITFOUNTTASK,
        "__module__": "bitfount.federated.transport.protos.messages_pb2"
        # @@protoc_insertion_point(class_scope:messages.BitfountTask)
    },
)
_sym_db.RegisterMessage(BitfountTask)

BitfountTasks = _reflection.GeneratedProtocolMessageType(
    "BitfountTasks",
    (_message.Message,),
    {
        "DESCRIPTOR": _BITFOUNTTASKS,
        "__module__": "bitfount.federated.transport.protos.messages_pb2"
        # @@protoc_insertion_point(class_scope:messages.BitfountTasks)
    },
)
_sym_db.RegisterMessage(BitfountTasks)

CommunicationDetails = _reflection.GeneratedProtocolMessageType(
    "CommunicationDetails",
    (_message.Message,),
    {
        "PodMailboxIdsEntry": _reflection.GeneratedProtocolMessageType(
            "PodMailboxIdsEntry",
            (_message.Message,),
            {
                "DESCRIPTOR": _COMMUNICATIONDETAILS_PODMAILBOXIDSENTRY,
                "__module__": "bitfount.federated.transport.protos.messages_pb2"
                # @@protoc_insertion_point(class_scope:messages.CommunicationDetails.PodMailboxIdsEntry)
            },
        ),
        "DESCRIPTOR": _COMMUNICATIONDETAILS,
        "__module__": "bitfount.federated.transport.protos.messages_pb2"
        # @@protoc_insertion_point(class_scope:messages.CommunicationDetails)
    },
)
_sym_db.RegisterMessage(CommunicationDetails)
_sym_db.RegisterMessage(CommunicationDetails.PodMailboxIdsEntry)

BitfountMessage = _reflection.GeneratedProtocolMessageType(
    "BitfountMessage",
    (_message.Message,),
    {
        "PodMailboxIdsEntry": _reflection.GeneratedProtocolMessageType(
            "PodMailboxIdsEntry",
            (_message.Message,),
            {
                "DESCRIPTOR": _BITFOUNTMESSAGE_PODMAILBOXIDSENTRY,
                "__module__": "bitfount.federated.transport.protos.messages_pb2"
                # @@protoc_insertion_point(class_scope:messages.BitfountMessage.PodMailboxIdsEntry)
            },
        ),
        "DESCRIPTOR": _BITFOUNTMESSAGE,
        "__module__": "bitfount.federated.transport.protos.messages_pb2"
        # @@protoc_insertion_point(class_scope:messages.BitfountMessage)
    },
)
_sym_db.RegisterMessage(BitfountMessage)
_sym_db.RegisterMessage(BitfountMessage.PodMailboxIdsEntry)

Acknowledgement = _reflection.GeneratedProtocolMessageType(
    "Acknowledgement",
    (_message.Message,),
    {
        "DESCRIPTOR": _ACKNOWLEDGEMENT,
        "__module__": "bitfount.federated.transport.protos.messages_pb2"
        # @@protoc_insertion_point(class_scope:messages.Acknowledgement)
    },
)
_sym_db.RegisterMessage(Acknowledgement)

SuccessResponse = _reflection.GeneratedProtocolMessageType(
    "SuccessResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _SUCCESSRESPONSE,
        "__module__": "bitfount.federated.transport.protos.messages_pb2"
        # @@protoc_insertion_point(class_scope:messages.SuccessResponse)
    },
)
_sym_db.RegisterMessage(SuccessResponse)

LargeStorageRequest = _reflection.GeneratedProtocolMessageType(
    "LargeStorageRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _LARGESTORAGEREQUEST,
        "__module__": "bitfount.federated.transport.protos.messages_pb2"
        # @@protoc_insertion_point(class_scope:messages.LargeStorageRequest)
    },
)
_sym_db.RegisterMessage(LargeStorageRequest)

LargeStorageTaskRequest = _reflection.GeneratedProtocolMessageType(
    "LargeStorageTaskRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _LARGESTORAGETASKREQUEST,
        "__module__": "bitfount.federated.transport.protos.messages_pb2"
        # @@protoc_insertion_point(class_scope:messages.LargeStorageTaskRequest)
    },
)
_sym_db.RegisterMessage(LargeStorageTaskRequest)

BlobStorageData = _reflection.GeneratedProtocolMessageType(
    "BlobStorageData",
    (_message.Message,),
    {
        "UploadFieldsEntry": _reflection.GeneratedProtocolMessageType(
            "UploadFieldsEntry",
            (_message.Message,),
            {
                "DESCRIPTOR": _BLOBSTORAGEDATA_UPLOADFIELDSENTRY,
                "__module__": "bitfount.federated.transport.protos.messages_pb2"
                # @@protoc_insertion_point(class_scope:messages.BlobStorageData.UploadFieldsEntry)
            },
        ),
        "DESCRIPTOR": _BLOBSTORAGEDATA,
        "__module__": "bitfount.federated.transport.protos.messages_pb2"
        # @@protoc_insertion_point(class_scope:messages.BlobStorageData)
    },
)
_sym_db.RegisterMessage(BlobStorageData)
_sym_db.RegisterMessage(BlobStorageData.UploadFieldsEntry)


_COMMUNICATIONDETAILS_PODMAILBOXIDSENTRY._options = None
_BITFOUNTMESSAGE_PODMAILBOXIDSENTRY._options = None
_BLOBSTORAGEDATA_UPLOADFIELDSENTRY._options = None

_MESSAGESERVICE = _descriptor.ServiceDescriptor(
    name="MessageService",
    full_name="messages.MessageService",
    file=DESCRIPTOR,
    index=0,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
    serialized_start=2084,
    serialized_end=2568,
    methods=[
        _descriptor.MethodDescriptor(
            name="PodConnect",
            full_name="messages.MessageService.PodConnect",
            index=0,
            containing_service=None,
            input_type=_PODDATA,
            output_type=_SUCCESSRESPONSE,
            serialized_options=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.MethodDescriptor(
            name="SetupTaskMailboxes",
            full_name="messages.MessageService.SetupTaskMailboxes",
            index=1,
            containing_service=None,
            input_type=_BITFOUNTTASKS,
            output_type=_COMMUNICATIONDETAILS,
            serialized_options=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.MethodDescriptor(
            name="AcknowledgeMessage",
            full_name="messages.MessageService.AcknowledgeMessage",
            index=2,
            containing_service=None,
            input_type=_ACKNOWLEDGEMENT,
            output_type=_SUCCESSRESPONSE,
            serialized_options=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.MethodDescriptor(
            name="GetBitfountMessage",
            full_name="messages.MessageService.GetBitfountMessage",
            index=3,
            containing_service=None,
            input_type=_COMMUNICATIONDETAILS,
            output_type=_BITFOUNTMESSAGE,
            serialized_options=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.MethodDescriptor(
            name="SendBitfountMessage",
            full_name="messages.MessageService.SendBitfountMessage",
            index=4,
            containing_service=None,
            input_type=_BITFOUNTMESSAGE,
            output_type=_SUCCESSRESPONSE,
            serialized_options=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.MethodDescriptor(
            name="GetLargeObjectStorage",
            full_name="messages.MessageService.GetLargeObjectStorage",
            index=5,
            containing_service=None,
            input_type=_LARGESTORAGEREQUEST,
            output_type=_BLOBSTORAGEDATA,
            serialized_options=None,
            create_key=_descriptor._internal_create_key,
        ),
    ],
)
_sym_db.RegisterServiceDescriptor(_MESSAGESERVICE)

DESCRIPTOR.services_by_name["MessageService"] = _MESSAGESERVICE

# @@protoc_insertion_point(module_scope)



"""Tests datastructure.py."""
import re
from typing import Dict, List, Union
from unittest.mock import MagicMock, Mock

import pandas as pd
import pytest
from pytest_asyncio import fixture

from bitfount.data.datasources.database_source import DatabaseSource
from bitfount.data.datasources.dataframe_source import DataFrameSource
from bitfount.data.datastructure import DEFAULT_IMAGE_TRANSFORMATIONS, DataStructure
from bitfount.data.exceptions import DataStructureError
from bitfount.data.types import DataSplit, SchemaOverrideMapping
from bitfount.runners.config_schemas import (
    DataStructureAssignConfig,
    DataStructureSelectConfig,
    DataStructureTableConfig,
    DataStructureTransformConfig,
)
from bitfount.transformations.batch_operations import AlbumentationsImageTransformation
from bitfount.types import _StrAnyDict
from tests.utils.helper import TABLE_NAME, create_schema, integration_test, unit_test


@unit_test
class TestDataStructure:
    """Tests DataStructure class."""

    @fixture
    def query(self) -> str:
        """Query fixture."""
        return "SELECT * from TABLE"

    def test_transformed_columns_added_to_selected_cols(self) -> None:
        """Tests that the transformed columns are added to selected_cols."""
        dataset: List[Dict[str, _StrAnyDict]] = [
            {"convert_to": {"type": "string", "col": "TARGET"}},
            {"normalizedata": {"col": ["A", "B"], "keep_original": "A"}},
        ]
        batch = [{"albumentations": {"arg": "col_1", "output": True}}]
        datastructure = DataStructure(
            target="TARGET",
            dataset_transforms=dataset,
            batch_transforms=batch,
            selected_cols=["A", "B", "C", "image"],
            table=TABLE_NAME,
        )
        assert set(datastructure.selected_cols) == set(
            ["A", "A_normalizedata", "B_normalizedata", "C", "image"]
        )

    def test_transformed_columns_added_to_ignored_cols(self) -> None:
        """Tests that the transformed columns are added to ignore_cols."""
        dataset: List[Dict[str, _StrAnyDict]] = [
            {"convert_to": {"type": "string", "col": "TARGET"}},
            {"normalizedata": {"col": ["A", "B"], "keep_original": "A"}},
        ]
        batch = [{"albumentations": {"arg": "col_1", "output": True}}]
        datastructure = DataStructure(
            target="TARGET",
            dataset_transforms=dataset,
            batch_transforms=batch,
            ignore_cols=["Date"],
            table=TABLE_NAME,
        )
        assert set(datastructure.ignore_cols) == set(["B", "Date"])

    def test_create_datastructure_from_config(self) -> None:
        """Tests that the datastructure gets the right args fom configs."""
        select = DataStructureSelectConfig(
            include=["TARGET", "weights", "Col1", "image"]
        )
        assign = DataStructureAssignConfig(
            target="TARGET",
            loss_weights_col="weights",
            ignore_classes_col="ignore_classes",
        )
        transform = DataStructureTransformConfig(
            dataset=[{"convert_to": {"type": "string", "col": "TARGET"}}],
            batch=[{"albumentations": {"col": ["image"], "keep_original": "image"}}],
        )
        table_config = DataStructureTableConfig(table=TABLE_NAME)
        ds = DataStructure.create_datastructure(
            select=select,
            transform=transform,
            assign=assign,
            table_config=table_config,
        )
        assert ds.target == "TARGET"
        assert ds.ignore_cols == []
        assert set(ds.selected_cols) == set(["TARGET", "weights", "Col1", "image"])
        assert ds.loss_weights_col == "weights"
        assert ds.ignore_classes_col == "ignore_classes"
        assert ds.batch_transforms == [
            {"albumentations": {"col": ["image"], "keep_original": "image"}}
        ]
        assert ds.dataset_transforms == [
            {"convert_to": {"type": "string", "col": "TARGET"}}
        ]

    def test_create_datastructure_select_config_raises_error(self) -> None:
        """Tests that the providing both include and exclude cols raises error."""
        select = DataStructureSelectConfig(
            include=["TARGET", "weights", "multihead_col", "Col1"],
            exclude=["Col2", "Col3"],
        )
        transform = DataStructureTransformConfig()
        assign = DataStructureAssignConfig(target="TARGET")
        table_config = DataStructureTableConfig(table=TABLE_NAME)
        with pytest.raises(DataStructureError):
            DataStructure.create_datastructure(
                select=select,
                transform=transform,
                assign=assign,
                table_config=table_config,
            )

    def test_invalid_batch_transformations(self) -> None:
        """Tests invalid batch transformations raise a `ValueError`."""
        batch = [
            {
                "invalid_transformation": {
                    "step": "train",
                    "output": True,
                    "arg": "col1",
                    "transformations": ["RandomBrightnessContrast"],
                }
            },
            {
                "another_invalid_transformation": {
                    "step": "train",
                    "output": True,
                    "arg": "col1",
                    "transformations": ["RandomBrightnessContrast"],
                }
            },
        ]
        with pytest.raises(
            ValueError,
            match=re.escape(
                "The following batch transformations are not recognised: "
                "another_invalid_transformation, invalid_transformation."
            ),
        ):
            DataStructure(
                target="TARGET",
                batch_transforms=batch,
                selected_cols=["A", "B", "C", "image"],
                table=TABLE_NAME,
            )

    def test_get_batch_transformations(self) -> None:
        """Tests `get_batch_transformations` method."""
        batch = [
            {
                "albumentations": {
                    "step": "train",
                    "output": True,
                    "arg": "col1",
                    "transformations": ["RandomBrightnessContrast"],
                }
            }
        ]
        ds = DataStructure(
            target="TARGET",
            batch_transforms=batch,
            selected_cols=["A", "B", "C", "image"],
            table=TABLE_NAME,
        )
        tfms = ds.get_batch_transformations()
        assert tfms is not None
        assert len(tfms) == 1
        assert isinstance(tfms[0], AlbumentationsImageTransformation)

    def test_setting_default_image_transformations(self) -> None:
        """Tests that default batch transformations are set if not provided."""
        ds = DataStructure(
            target="TARGET",
            selected_cols=["A", "B", "C", "image"],
            image_cols=["image"],
            table=TABLE_NAME,
        )
        assert ds.batch_transforms is not None
        assert ds.batch_transforms == [
            {
                "albumentations": {
                    "arg": "image",
                    "output": True,
                    "transformations": DEFAULT_IMAGE_TRANSFORMATIONS,
                    "step": "train",
                }
            },
            {
                "albumentations": {
                    "arg": "image",
                    "output": True,
                    "transformations": DEFAULT_IMAGE_TRANSFORMATIONS,
                    "step": "validation",
                }
            },
            {
                "albumentations": {
                    "arg": "image",
                    "output": True,
                    "transformations": DEFAULT_IMAGE_TRANSFORMATIONS,
                    "step": "test",
                }
            },
        ]

    def test_get_dataset_transformations_with_default_batch_transformations(
        self,
    ) -> None:
        """Tests that the default batch transformations are loaded correctly."""
        ds = DataStructure(
            target="TARGET",
            selected_cols=["A", "B", "C", "image"],
            image_cols=["image"],
            table=TABLE_NAME,
        )
        tfms = ds.get_batch_transformations()
        assert tfms is not None
        assert len(tfms) == 3
        for tfm, split in zip(tfms, DataSplit):
            assert isinstance(tfm, AlbumentationsImageTransformation)
            assert tfm.step == split

    def test_empty_batch_transformations(self) -> None:
        """Tests that batch transformations can be an empty list.

        Tests that if an empty list is explicitly provided and there are no image
        transformations, then the batch transformations stay empty.
        """
        ds = DataStructure(
            target="TARGET",
            selected_cols=["A", "B", "C", "image"],
            batch_transforms=[],
            table=TABLE_NAME,
        )
        assert ds.batch_transforms == []
        assert ds.get_batch_transformations() == []

    def test_get_table_name_with_single_table_datastructure(self) -> None:
        """Tests that the table name is returned correctly."""
        ds = DataStructure(
            target="TARGET",
            table=TABLE_NAME,
        )
        assert ds.get_table_name() == TABLE_NAME

    def test_get_table_name_with_multi_pod_datastructure(self) -> None:
        """Tests that the table name is returned correctly."""
        ds = DataStructure(
            target="TARGET",
            table={"pod1": "pod1_table", "pod2": "pod2_table"},
        )
        assert ds.get_table_name("pod1") == "pod1_table"

    def test_get_table_name_with_multi_pod_datastructure_raises_value_error(
        self,
    ) -> None:
        """Tests that the `get_table_name` method raises a ValueError."""
        ds = DataStructure(
            target="TARGET",
            table={"pod1": "pod1_table", "pod2": "pod2_table"},
        )
        with pytest.raises(
            ValueError, match="No pod identifier provided for multi-pod datastructure."
        ):
            ds.get_table_name()

    @pytest.mark.parametrize(
        "query", ["SELECT * FROM blah", {"pod1": "SELECT * FROM blah"}]
    )
    def test_get_table_schema_query(
        self, mocker: Mock, query: Union[str, dict]
    ) -> None:
        """Test get_table_schema with query."""
        mock_override_schema = mocker.patch.object(DataStructure, "_override_schema")
        mock_datasource = MagicMock(spec=DatabaseSource, query="SELECT * FROM table")
        mock_datasource.__len__.return_value = 1000
        pod_id = "pod1"

        schema_override: SchemaOverrideMapping = {
            "categorical": [{"col1": {"value_1": 0, "value_2": 1}}],
            "continuous": ["col2"],
        }

        pod_schema_override = {pod_id: schema_override}

        ds = DataStructure(
            target="TARGET", query=query, schema_types_override=pod_schema_override
        )
        ds.get_table_schema(
            schema=Mock(), pod_identifier=pod_id, datasource=mock_datasource
        )

        mock_override_schema.assert_called_once()

    def test_get_pod_identifiers_with_single_table_datastructure(self) -> None:
        """Tests that `get_pod_identifiers` returns None if there are no pods."""
        ds = DataStructure(
            target="TARGET",
            table=TABLE_NAME,
        )
        assert ds.get_pod_identifiers() is None

    def test_get_pod_identifiers_with_multi_pod_datastructure(self) -> None:
        """Tests that `get_pod_identifiers` returns the correct identifiers."""
        ds = DataStructure(
            target="TARGET",
            table={"pod1": "pod1_table", "pod2": "pod2_table"},
        )
        assert ds.get_pod_identifiers() == ["pod1", "pod2"]

    def test_error_raised_for_incompatible_parameters(self) -> None:
        """Tests error is raised when incompatible parameters are given."""
        selected_cols = ["a", "b"]
        ignore_cols = ["c", "d"]
        with pytest.raises(DataStructureError):
            DataStructure(
                target="TARGET",
                table=TABLE_NAME,
                selected_cols=selected_cols,
                ignore_cols=ignore_cols,
            )

    def test_both_table_and_query_error(self, query: str) -> None:
        """Tests error is raised when both table and query are given."""
        with pytest.raises(DataStructureError):
            DataStructure(
                target="TARGET",
                query=query,
                table=TABLE_NAME,
            )

    def test_both_query_no_schema_override_error(self, query: str) -> None:
        """Tests error is raised when query is given with no schema override."""
        with pytest.raises(DataStructureError):
            DataStructure(
                target="TARGET",
                query=query,
            )

    def test_both_query_selected_cols_error(self, query: str) -> None:
        """Tests error is raised when both selected_cols and query are given."""
        schema_override: SchemaOverrideMapping = {
            "categorical": [{"col1": {"value_1": 0, "value_2": 1}}],
            "continuous": ["col2"],
        }
        with pytest.raises(DataStructureError):
            DataStructure(
                target="TARGET",
                query=query,
                schema_types_override=schema_override,
                selected_cols=["a"],
            )

    def test_both_query_ignore_cols_error(self, query: str) -> None:
        """Tests error is raised when both ignore_cols and query are given."""
        schema_override: SchemaOverrideMapping = {
            "categorical": [{"col1": {"value_1": 0, "value_2": 1}}],
            "continuous": ["col2"],
        }
        with pytest.raises(DataStructureError):
            DataStructure(
                target="TARGET",
                query=query,
                schema_types_override=schema_override,
                ignore_cols=["a"],
            )

    def test_schema_overide_local_categorical_no_encodings_error(
        self, query: str
    ) -> None:
        """Tests error raised if schema override has categorical attr but no encoding.

        Test for datastructure meant for local training.
        """
        schema_override: SchemaOverrideMapping = {
            "categorical": ["col1"],
            "continuous": ["col2"],
        }
        with pytest.raises(DataStructureError):
            DataStructure(
                target="TARGET",
                query=query,
                schema_types_override=schema_override,
            )

    def test_schema_overide_remote_categorical_no_encodings_error(
        self, query: str
    ) -> None:
        """Tests error raised if schema override has categorical attr but no encoding.

        Test for datastructure meant for remote training.
        """
        schema_override: Dict[str, SchemaOverrideMapping] = {
            "pod_id": {
                "categorical": ["col1"],
                "continuous": ["col2"],
            },
        }
        with pytest.raises(DataStructureError):
            DataStructure(
                target="TARGET",
                query=query,
                schema_types_override=schema_override,
                ignore_cols=["a"],
            )

    def test_datastructure_no_query_no_table_error(self) -> None:
        """Tests error is raised with both query and table."""
        with pytest.raises(DataStructureError):
            DataStructure(
                target="TARGET",
            )

    def test_get_pod_identifiers_with_query_local_datastructure(
        self, query: str
    ) -> None:
        """Tests that `get_pod_identifiers` returns None if there are no pods."""
        schema_override: SchemaOverrideMapping = {
            "categorical": [{"col1": {"value_1": 0, "value_2": 1}}],
            "continuous": ["col2"],
        }
        ds = DataStructure(
            target="TARGET", query=query, schema_types_override=schema_override
        )
        assert ds.get_pod_identifiers() is None

    def test_get_pod_identifiers_with_multi_pod_query(self, query: str) -> None:
        """Tests that `get_pod_identifiers` returns the correct identifiers."""
        schema_override: Dict[str, SchemaOverrideMapping] = {
            "pod1": {
                "categorical": [{"col1": {"value_1": 0, "value_2": 1}}],
                "continuous": ["col2"],
            },
            "pod2": {
                "categorical": [{"col1": {"value_1": 0, "value_2": 1}}],
                "continuous": ["col2"],
            },
        }
        ds = DataStructure(
            target="TARGET",
            query={"pod1": query, "pod2": query},
            schema_types_override=schema_override,
        )
        assert ds.get_pod_identifiers() == ["pod1", "pod2"]

    def test_query_pod_id_not_in_override(self, query: str) -> None:
        """Tests that pod_id not found in override raises error."""
        schema_override: SchemaOverrideMapping = {
            "categorical": [{"col1": {"value_1": 0, "value_2": 1}}],
            "continuous": ["col2"],
        }
        with pytest.raises(DataStructureError):
            DataStructure(
                target="TARGET",
                query={"pod1": query, "pod2": query},
                schema_types_override=schema_override,
            )

    def test_query_pod_id_no_encodings(self, query: str) -> None:
        """Tests that pod_id not found in override raises error."""
        schema_override: Dict[str, SchemaOverrideMapping] = {
            "pod1": {
                "categorical": ["col1"],
                "continuous": ["col2"],
            },
            "pod2": {
                "categorical": ["col1"],
                "continuous": ["col2"],
            },
        }
        with pytest.raises(DataStructureError):
            DataStructure(
                target="TARGET",
                query={"pod1": query, "pod2": query},
                schema_types_override=schema_override,
            )

    def test_override_schema_raises_value_error(self) -> None:
        """Test that the schema overrides raises a ValueError if no query."""
        ds = DataStructure(target="TARGET", table="table")
        with pytest.raises(
            ValueError,
            match="No query or dictionary of pod_identifiers to queries was given.",
        ):
            ds._override_schema()

    def test_override_schema_local(self, query: str) -> None:
        """Test that the schema overrides returns the expected features."""
        schema_override: SchemaOverrideMapping = {
            "categorical": [{"col1": {"value_1": 0, "value_2": 1}}],
            "continuous": ["col2"],
            "text": ["col3"],
            "image": ["col4"],
        }
        ds = DataStructure(
            target="TARGET", query=query, schema_types_override=schema_override
        )
        schema = ds._override_schema()
        assert "col1" in schema.features["categorical"]
        assert schema.features["categorical"]["col1"].encoder.size == 2
        assert "col2" in schema.features["continuous"]
        assert "col3" in schema.features["text"]
        assert "col4" in schema.features["image"]

    def test_override_schema_pod_id(self, query: str) -> None:
        """Test that the schema overrides returns the expected features."""
        pod_id = "pod_id"
        schema_override: Dict[str, SchemaOverrideMapping] = {
            pod_id: {
                "categorical": [{"col1": {"value_1": 0, "value_2": 1}}],
                "continuous": ["col2"],
                "text": ["col3"],
                "image": ["col4"],
            }
        }
        ds = DataStructure(
            target="TARGET",
            query={pod_id: query},
            schema_types_override=schema_override,
        )
        schema = ds._override_schema(pod_identifier=pod_id)
        assert "col1" in schema.features["categorical"]
        assert schema.features["categorical"]["col1"].encoder.size == 2
        assert "col2" in schema.features["continuous"]
        assert "col3" in schema.features["text"]
        assert "col4" in schema.features["image"]

    def test_override_schema_datasource(self, query: str) -> None:
        """Test that the schema overrides returns the expected features."""
        pod_id = "pod_id"
        schema_override: Dict[str, SchemaOverrideMapping] = {
            pod_id: {
                "categorical": [{"col1": {"value_1": 0, "value_2": 1}}],
                "continuous": ["col2"],
                "text": ["col3"],
                "image": ["col4"],
            }
        }
        ds = DataStructure(
            target="TARGET",
            query={pod_id: query},
            schema_types_override=schema_override,
        )
        data = {
            "col1": ["value_1", "value_2", "value_2"],
            "col2": [0, 1, 2],
            "col3": ["str1", "str2", "str3"],
            "col4": ["img1", "img2", "img3"],
            "col5": [0.2, 3, 24],
        }
        datasource = DataFrameSource(pd.DataFrame(data=data))
        datasource.load_data()
        schema = ds._override_schema(datasource, pod_id)
        assert "col1" in schema.features["categorical"]
        assert schema.features["categorical"]["col1"].encoder.size == 2
        assert "col2" in schema.features["continuous"]
        assert "col3" in schema.features["text"]
        assert "col4" in schema.features["image"]
        assert "col5" not in schema.features

    def test_update_datastructure_with_hub_identifiers_query(self, query: str) -> None:
        """Tests that the pod identifiers get updated."""
        pod_id = "pod_id"
        schema_override: Dict[str, SchemaOverrideMapping] = {
            pod_id: {
                "categorical": [{"col1": {"value_1": 0, "value_2": 1}}],
                "continuous": ["col2"],
                "text": ["col3"],
                "image": ["col4"],
            }
        }
        ds = DataStructure(
            target="TARGET",
            query={pod_id: query},
            schema_types_override=schema_override,
        )
        updated_pod_id = "testing/pod_id"
        ds._update_datastructure_with_hub_identifiers([updated_pod_id])
        assert isinstance(ds.query, dict)
        assert updated_pod_id in ds.query.keys()
        assert isinstance(ds.schema_types_override, dict)
        assert updated_pod_id in ds.schema_types_override.keys()


@integration_test
def test_get_columns_ignored_for_training() -> None:
    """Tests that `ignore_cols` is setup correctly from schema and `selected_cols`."""
    schema = create_schema(classification=True)
    datastructure = DataStructure(
        target="TARGET",
        selected_cols=[
            "A",
            "B",
            "C",
            "D",
            "E",
            "F",
            "G",
            "H",
            "I",
            "J",
            "K",
            "L",
            "M",
            "N",
        ],
        table=TABLE_NAME,
    )
    table_schema = datastructure.get_table_schema(
        schema=schema,
    )
    datastructure.get_columns_ignored_for_training(table_schema)
    assert set(datastructure.ignore_cols) == set(["O", "P", "Date", "TARGET"])



"""Test secure multi party computation with pytorch backend."""
import copy
from functools import partial
from typing import Any, Callable, Dict, List, Tuple, cast
from unittest.mock import AsyncMock, MagicMock, create_autospec

from _pytest.logging import LogCaptureFixture
import numpy as np
import pandas as pd
import pytest
from pytest import fixture
from pytest_mock import MockerFixture
import torch

from bitfount.backends.pytorch.models.models import PyTorchTabularClassifier
from bitfount.backends.pytorch.types import _AdaptorForPyTorchTensor
from bitfount.data.datasources.base_source import BaseSource
from bitfount.data.datastructure import DataStructure
from bitfount.data.schema import BitfountSchema
from bitfount.federated.secure import SecureShare
from bitfount.federated.transport.worker_transport import _WorkerMailbox
from bitfount.types import _TensorLike, _Weights
from tests.utils.helper import (
    backend_test,
    create_dataset,
    create_datasource,
    create_datastructure,
    unit_test,
)


@backend_test
@unit_test
class TestSecureShare:
    """Test SecureShare algorithm with pytorch tensors."""

    @fixture
    def datastructure(self) -> DataStructure:
        """Fixture for datastructure."""
        return create_datastructure()

    @fixture
    def datasource(self) -> BaseSource:
        """Fixture for datasource."""
        return create_datasource(classification=True)

    @fixture
    def mock_secure_transport_methods(self) -> Tuple[Callable, Callable]:
        """Returns tuple of mocked SecureShare methods which need a mailbox."""

        def mock_share_own_shares(worker_shares: Any, self: Any, mailbox: Any) -> None:
            """Mocks '_share_own_shares' method."""
            for worker, shares in worker_shares.items():
                if worker != self.name:
                    rand_num = self._get_random_number()
                    shares.append(rand_num)

        def mock_receive_worker_shares(
            worker_shares: Any, self: Any, mailbox: Any
        ) -> None:
            """Mocks '_receive_worker_shares' method."""
            self._other_worker_shares = worker_shares[self.name]
            worker_shares[self.name] = []

        return mock_share_own_shares, mock_receive_worker_shares

    @pytest.mark.parametrize(
        "prime_q, tensor_list, precision",
        [
            (5, [0.1, 0.2, -0.2], 10),
            (13, [0.1, -0.2, -0.3], 10),
            (29, [-1.0, 0.2, 0.9], 10),
            (101, [0.1, -0.39, 0.5], 100),
            (1913, [0.1, -4, 0.39], 100),
            (1913, [0.1, -0.4, 0.4], 1000),
        ],
    )
    def test_encode_and_decode_finite_field(
        self, precision: int, prime_q: int, tensor_list: List[float]
    ) -> None:
        """Ensures output matches input after encoding and decoding."""
        sec = SecureShare(prime_q=prime_q, precision=precision)
        tensor = _AdaptorForPyTorchTensor(
            torch.tensor(tensor_list, dtype=torch.float64)
        )
        encoded_tensor = sec._encode_finite_field(tensor)
        assert isinstance(encoded_tensor, np.ndarray)
        decoded_tensor = sec._decode_finite_field(encoded_tensor)
        assert isinstance(decoded_tensor, np.ndarray)
        np.testing.assert_array_equal(tensor.torchtensor.numpy(), decoded_tensor)

    @pytest.mark.parametrize(
        "a, b",
        [
            ([13, 10], [29, 1]),
            ([13, 50], [49, 7]),
            ([-13, -50], [-49, -7]),
            ([-126, -41], [109, 111]),
            ([-123, 50], [-1, 17]),
            ([-123.23, 50.1], [-1.01, 16.99]),
            ([0.1234001, 0.23451], [0.13245, 0.674532]),
        ],
    )
    def test_addition_finite_field(self, a: List[float], b: List[float]) -> None:
        """Tests addition of encoded tensors, compares result with regular addition."""
        sec = SecureShare()
        encoded_a = sec._encode_finite_field(
            _AdaptorForPyTorchTensor(torch.tensor(a, dtype=torch.float32))
        )
        encoded_b = sec._encode_finite_field(
            _AdaptorForPyTorchTensor(torch.tensor(b, dtype=torch.float32))
        )
        encoded_result = sec._add([encoded_a, encoded_b])
        decoded_result = sec._decode_finite_field(encoded_result)

        np.testing.assert_almost_equal(
            np.array(a) + np.array(b), decoded_result, decimal=4
        )

    @pytest.mark.parametrize(
        "prime_q, tensor, error, num_other_workers",
        [
            (5, [0.1, 0.2, 0.3], True, 0),
            (7, [0.1, -0.2, -0.3], False, 0),
            (7, [0.1, 0.2, 0.3], False, 0),
            (7, [0.1, -0.39, 0.39], False, 0),
            (7, [0.1, -0.4, 0.39], True, 0),
            (7, [0.1, -0.39, 0.4], True, 0),
            (13, [0.1, -0.2, -0.3], False, 1),
            (13, [0.1, 0.2, 0.3], True, 2),
            (13, [0.1, -0.39, 0.39], True, 3),
        ],
    )
    def test_encode_finite_field_secure_share_error(
        self,
        caplog: LogCaptureFixture,
        error: bool,
        num_other_workers: int,
        prime_q: int,
        tensor: List[float],
    ) -> None:
        """Tests that clamping is applied during encoding.

        When there are not enough positive integers in prime_q to map every positive
        and negative integer to an integer in the finite field without multiple values
        corresponding to the same value, we clamp the parameters before encoding.
        This is applied *only* when steps between parameter updates in the protocol
        is set to 1.
        """
        sec = SecureShare(prime_q=prime_q, precision=10)
        sec._own_shares = [1]
        sec._own_shares = sec._own_shares * num_other_workers
        if error:
            sec._encode_finite_field(_AdaptorForPyTorchTensor(torch.tensor(tensor)))
            assert (
                caplog.records[0].msg
                == "Parameter weights have been clipped. If you want to avoid this, "
                "choose a larger `prime_q` or a smaller `precision` for "
                "the `SecureShare` or normalize continuous features prior "
                "to training."
            )
        else:
            sec._encode_finite_field(_AdaptorForPyTorchTensor(torch.tensor(tensor)))

    def test_get_random_number(self) -> None:
        """Tests _get_random_number method."""
        sec = SecureShare(prime_q=101, precision=10)
        assert sec._own_shares == []
        for i in range(1, 10):
            sec._get_random_number()
            assert len(sec._own_shares) == i
            assert sec._own_shares[-1] < sec.prime_q

    @pytest.mark.parametrize(
        "input, output",
        [
            ([0.6, 0.7, 0.8], [1, 2, 3]),
            ([-1, -4, -3], [994, 964, 974]),
            ([-2, -3, 4.1], [984, 974, 36]),
        ],
    )
    def test_encode_secret(self, input: List[float], output: List[int]) -> None:
        """Tests _encode_secret method with fixed shares and simple tensors."""
        sec = SecureShare(prime_q=1009, precision=10)
        sec._own_shares = [-2, 3, 4]
        encoded_secret = sec._encode_secret(
            _AdaptorForPyTorchTensor(torch.tensor(input))
        )
        assert isinstance(encoded_secret, np.ndarray)
        np.testing.assert_array_equal(encoded_secret, np.array(output))

    def test_reconstruct_secret(self) -> None:
        """Tests _reconstruct_secret with simple tensors."""
        sec = SecureShare(prime_q=101, precision=10)
        reconstructed = sec._reconstruct_secret([np.array([1, 2, 3]), 4, 5, 6])
        assert isinstance(reconstructed, np.ndarray)
        assert len(reconstructed) == 3
        np.testing.assert_array_equal(reconstructed, np.array([16, 17, 18]))

    def test_encode_and_reconstruct_state_dict(self) -> None:
        """Tests _encode_and_reconstruct_state_dict with simple tensors."""
        sec = SecureShare()
        state_dict: _Weights = {
            "tensor_a": _AdaptorForPyTorchTensor(torch.tensor([-0.1, 0.9, 1.9])),
            "tensor_b": _AdaptorForPyTorchTensor(torch.tensor([-1.0, 0.0, 1.0])),
        }
        encoded_state_dict = sec._encode_and_reconstruct_state_dict(state_dict)
        assert isinstance(encoded_state_dict, dict)
        assert len(encoded_state_dict) == 2
        for param in encoded_state_dict.values():
            assert isinstance(param, np.ndarray)
            assert len(param) == 3

    async def test_do_secure_aggregation(self, mocker: MockerFixture) -> None:
        """Tests do_secure_aggregation method.

        Mocks out transport layer and makes the relevant assertions on calls/awaits.
        """
        sec = SecureShare()
        state_dict: Dict[str, _TensorLike] = {
            "tensor_a": _AdaptorForPyTorchTensor(torch.tensor([-0.1, 0.9, 1.9])),
            "tensor_b": _AdaptorForPyTorchTensor(torch.tensor([-1.0, 0.0, 1.0])),
        }
        num_other_pods = 3
        mock_mailbox = create_autospec(_WorkerMailbox, instance=True)

        # Create mocks for transport layer send and receive functions.
        def mock_send(secure_share_generator: Any, _mailbox: Any) -> None:
            """Mocks `_send_secure_shares_to_others` transport layer function."""
            for _ in range(num_other_pods):
                # As the result of this callable is stored on `SecureShare.own_shares`
                # we have to mock out the calls to it here.
                secure_share_generator()

        secure_share_send_mock: MagicMock = mocker.patch(
            "bitfount.federated.secure._send_secure_shares_to_others"
        )
        secure_share_send_mock.side_effect = mock_send
        secure_share_receive_mock: AsyncMock = mocker.patch(
            "bitfount.federated.secure._get_worker_secure_shares"
        )
        secure_share_receive_mock.return_value = [1, 2, 3]

        await sec.do_secure_aggregation(state_dict, mock_mailbox)

        # Assertions
        assert secure_share_send_mock.call_count == 1
        # This should only be awaited once as it returns all shares
        assert secure_share_receive_mock.await_count == 1
        # We should have generated one share per other pod
        assert len(sec._own_shares) == num_other_pods
        # We should have received one share from each other pod
        assert len(sec._other_worker_shares) == num_other_pods

    def test_average_and_decode_state_dicts(self) -> None:
        """Tests `average_and_decode_state_dicts` works correctly."""
        state_dict_a = {
            "tensor_a": torch.tensor([-0.9, 0.1, 1.1]),
            "tensor_b": torch.tensor([-2.0, -1.0, 0.0]),
        }
        state_dict_b = {
            "tensor_a": torch.tensor([-0.1, 0.9, 1.9]),
            "tensor_b": torch.tensor([-1.0, 0.0, 1.0]),
        }
        sec_a = SecureShare(prime_q=1009, precision=10)
        sec_b = SecureShare(prime_q=1009, precision=10)
        rand_a = sec_a._get_random_number()
        rand_b = sec_b._get_random_number()
        sec_a._other_worker_shares.append(rand_b)
        sec_b._other_worker_shares.append(rand_a)
        encoded_state_dict_a = sec_a._encode_and_reconstruct_state_dict(
            cast(_Weights, state_dict_a)
        )
        encoded_state_dict_b = sec_b._encode_and_reconstruct_state_dict(
            cast(_Weights, state_dict_b)
        )

        average_state_dict = sec_a.average_and_decode_state_dicts(
            [encoded_state_dict_a, encoded_state_dict_b],
        )

        expected_average = [
            torch.tensor([-0.5, 0.5, 1.5]),
            torch.tensor([-1.5, -0.5, 0.5]),
        ]

        for x, y in zip(average_state_dict.values(), expected_average):
            np.testing.assert_array_equal(x, y)

    @pytest.mark.parametrize(
        "a, b",
        [
            ([13, 10], [29, 1]),
            ([13, 50], [49, 7]),
            ([-13, -50], [-49, -7]),
            ([-126, -41], [109, 111]),
            ([-123, 50], [-1, 17]),
            ([-123.23, 50.1], [-1.01, 16.99]),
            ([0.1234001, 0.23451], [0.13245, 0.674532]),
        ],
    )
    def test_secure_averaging_with_simple_tensors_without_mailbox(
        self, a: List[float], b: List[float]
    ) -> None:
        """Tests secure averaging with simple tensors.

        The methods concerning the sharing and receiving of shares have been bypassed.
        """
        sec_a = SecureShare()
        sec_b = SecureShare()
        tensor_a = _AdaptorForPyTorchTensor(torch.tensor(a, dtype=torch.float64))
        tensor_b = _AdaptorForPyTorchTensor(torch.tensor(b, dtype=torch.float64))
        rand_a = sec_a._get_random_number()
        rand_b = sec_b._get_random_number()
        sec_a._other_worker_shares.append(rand_b)
        sec_b._other_worker_shares.append(rand_a)
        encoded_tensor_a = sec_a._encode_and_reconstruct_state_dict(
            {"tensor": tensor_a}
        )
        encoded_tensor_b = sec_b._encode_and_reconstruct_state_dict(
            {"tensor": tensor_b}
        )

        encoded_result = sec_a._add(
            [encoded_tensor_a["tensor"], encoded_tensor_b["tensor"]]
        )
        decoded_result = sec_a._decode_finite_field(encoded_result)

        np.testing.assert_almost_equal(
            np.array(a) + np.array(b), decoded_result, decimal=4
        )

    @pytest.mark.parametrize("num_workers", [2, 3, 5])
    async def test_secure_averaging_with_simple_tensors(
        self,
        mock_secure_transport_methods: Tuple[Callable, Callable],
        mocker: MockerFixture,
        num_workers: int,
    ) -> None:
        """Tests secure averaging with simple tensors.

        The methods which rely on the transport layer have been mocked.
        """
        worker_shares: Dict[str, List[SecureShare]] = {}
        nn_params: List[_Weights] = []
        for i in range(num_workers):
            params = {
                "tensor": torch.tensor([-5.1, -3.2, -1.3], dtype=torch.float64) + i
            }
            nn_params.append(cast(Dict[str, _TensorLike], params))
            worker_shares[f"worker_{i}"] = []

        mocker.patch.object(
            SecureShare,
            "_share_own_shares",
            autospec=True,
            side_effect=partial(mock_secure_transport_methods[0], worker_shares),
        )
        mocker.patch.object(
            SecureShare,
            "_receive_worker_shares",
            autospec=True,
            side_effect=partial(mock_secure_transport_methods[1], worker_shares),
        )

        secures: List[SecureShare] = []

        for name in worker_shares:
            sec = SecureShare()
            sec.name = name  # type: ignore[attr-defined] # Reason: name attr is only used in our patched functions above # noqa: B950
            await sec._share_own_shares(None)  # type: ignore[arg-type] # Reason: the function is mocked # noqa: B950
            secures.append(sec)

        for sec in secures:
            await sec._receive_worker_shares(None)  # type: ignore[arg-type] # Reason: the function is mocked # noqa: B950

        encoded_weights = []

        for i, sec in enumerate(secures):
            reconstructed_secret = list(
                sec._encode_and_reconstruct_state_dict(nn_params[i]).values()
            )
            encoded_weights.append(reconstructed_secret[0])

        summed_weights = encoded_weights[0]
        for i in range(1, len(encoded_weights)):
            summed_weights = sec._add([summed_weights, encoded_weights[i]])

        decoded: torch.Tensor = torch.tensor(sec._decode_finite_field(summed_weights))
        insecure_sum: torch.Tensor = torch.sum(
            torch.stack([cast(torch.Tensor, list(p.values())[0]) for p in nn_params]),
            dim=0,
        )
        for decoded_item, insecure_item in zip(decoded, insecure_sum):
            assert torch.isclose(decoded_item, insecure_item, atol=1e-4)

    @pytest.mark.parametrize("num_workers", [2, 3, 10])
    async def test_secure_averaging_with_neural_network_parameters(
        self,
        datasource: BaseSource,
        datastructure: DataStructure,
        mock_secure_transport_methods: Tuple[Callable, Callable],
        mocker: MockerFixture,
        num_workers: int,
    ) -> None:
        """Tests secure averaging with neural network parameters.

        The methods which rely on a transport layer have been mocked.
        """
        worker_shares: Dict[str, List[SecureShare]] = {}
        nn_params = []
        for i in range(num_workers):
            nn = PyTorchTabularClassifier(
                datastructure=datastructure, schema=BitfountSchema(), seed=i, epochs=2
            )
            nn.initialise_model(datasource)
            params = copy.deepcopy(nn.get_param_states())
            nn_params.append(nn._get_torch_tensor_states(params))
            worker_shares[f"worker_{i}"] = []
        weight_update = nn_params[0]
        for i in range(1, len(nn_params)):
            for param in nn_params[i]:
                weight_update[param] = weight_update[param] + nn_params[i][param]

        for param in weight_update:
            weight_update[param] = weight_update[param] / len(nn_params)

        mocker.patch.object(
            SecureShare,
            "_share_own_shares",
            autospec=True,
            side_effect=partial(mock_secure_transport_methods[0], worker_shares),
        )
        mocker.patch.object(
            SecureShare,
            "_receive_worker_shares",
            autospec=True,
            side_effect=partial(mock_secure_transport_methods[1], worker_shares),
        )

        secures: List[SecureShare] = []

        for name in worker_shares:
            sec = SecureShare()
            sec.name = name  # type: ignore[attr-defined] # Reason: name attr is only used in our patched functions above # noqa: B950
            await sec._share_own_shares(None)  # type: ignore[arg-type] # Reason: the function is mocked # noqa: B950
            secures.append(sec)

        for sec in secures:
            await sec._receive_worker_shares(None)  # type: ignore[arg-type] # Reason: the function is mocked # noqa: B950

        encoded_weights: List[Dict[str, np.ndarray]] = []

        for i, sec in enumerate(secures):
            reconstructed_state_dict: Dict[
                str, np.ndarray
            ] = sec._encode_and_reconstruct_state_dict(
                nn._get_torch_adapter_states(nn_params[i])
            )
            encoded_weights.append(reconstructed_state_dict)

        # It doesn't matter which SecureShare we use to average and decode the
        # state dictionaries so we just use the last one
        secure_decoded_average = sec.average_and_decode_state_dicts(encoded_weights)

        for param in nn_params[0]:
            insecure_average = nn_params[0][param]
            for i in range(1, len(nn_params)):
                insecure_average += nn_params[i][param]
            insecure_average /= len(nn_params)

            np.testing.assert_array_almost_equal(
                torch.mean(insecure_average).numpy(),
                np.mean(secure_decoded_average[param]),
                decimal=4,
            )

    @pytest.mark.parametrize("num_workers", [2, 3, 10])
    async def test_secure_averaging_with_pandas_dataframes(
        self,
        mock_secure_transport_methods: Tuple[Callable, Callable],
        mocker: MockerFixture,
        num_workers: int,
    ) -> None:
        """Tests secure averaging with neural network parameters.

        The methods which rely on a transport layer have been mocked.
        """
        worker_shares: Dict[str, List[SecureShare]] = {}
        dataframes: List[pd.DataFrame] = []
        for i in range(num_workers):
            # Using the index of each worker as the seed to ensure each dataset
            # is different. The dataset is also subsetted for simplicity.
            dataframes.append(create_dataset(seed=i)[["A", "B", "C", "D"]])
            worker_shares[f"worker_{i}"] = []

        mocker.patch.object(
            SecureShare,
            "_share_own_shares",
            autospec=True,
            side_effect=partial(mock_secure_transport_methods[0], worker_shares),
        )
        mocker.patch.object(
            SecureShare,
            "_receive_worker_shares",
            autospec=True,
            side_effect=partial(mock_secure_transport_methods[1], worker_shares),
        )

        secures: List[SecureShare] = []

        for name in worker_shares:
            sec = SecureShare()
            sec.name = name  # type: ignore[attr-defined] # Reason: name attr is only used in our patched functions above # noqa: B950
            await sec._share_own_shares(None)  # type: ignore[arg-type] # Reason: the function is mocked # noqa: B950
            secures.append(sec)

        for sec in secures:
            await sec._receive_worker_shares(None)  # type: ignore[arg-type] # Reason: the function is mocked # noqa: B950

        encoded_dataframes: List[Dict[str, np.ndarray]] = []

        for i, sec in enumerate(secures):
            reconstructed_state_dict = sec._encode_and_reconstruct_state_dict(
                {col: dataframes[i][col].to_numpy() for col in dataframes[i].columns}
            )
            encoded_dataframes.append(reconstructed_state_dict)

        sec = SecureShare()
        secure_decoded_average = sec.average_and_decode_state_dicts(encoded_dataframes)

        for field in dataframes[0].columns:
            insecure_average = dataframes[0][field]
            for i in range(1, len(dataframes)):
                insecure_average += dataframes[i][field]
            insecure_average /= len(dataframes)

            np.testing.assert_array_almost_equal(
                np.mean(insecure_average),
                np.mean(secure_decoded_average[field]),
                decimal=4,
            )



"""Types related to transport layer sending and receiving."""
from __future__ import annotations

import dataclasses
from dataclasses import dataclass
from datetime import datetime
from typing import Dict, NamedTuple, Type, TypeVar

from bitfount.types import _JSONDict

_T = TypeVar("_T")


@dataclass
class _SerializationMixin:
    """Mixin to provide (de)serialization functionality."""

    def serialize(self) -> _JSONDict:
        """Serialize the dataclass to a JSON-compatible dictionary."""
        return dataclasses.asdict(self)

    @classmethod
    def deserialize(cls: Type[_T], data: _JSONDict) -> _T:
        """Deserialize the dataclass from a JSON-compatible dictionary."""
        return cls(**data)


@dataclass
class _OIDCClientID(_SerializationMixin):
    """Form of Client ID response from pods to initial request."""

    client_id: str


@dataclass
class _OIDCAuthEndpointResponse:
    """Form of the details extracted from OIDC Auth Flow callback."""

    auth_code: str
    state: str


@dataclass
class _OIDCAuthFlowResponse(_SerializationMixin):
    """Form of OIDC Auth Flow message sent to pod for verification."""

    auth_code: str
    code_verifier: str
    redirect_uri: str


class _DeviceCodeDetailsPair(NamedTuple):
    """The separated pod- and modeller-related device code response details."""

    pod_details: _PodDeviceCodeDetails
    modeller_details: _ModellerDeviceCodeDetails


@dataclass
class _PodDeviceCodeDetails(_SerializationMixin):
    """The pod-related device code response details."""

    device_code: str
    expires_at: datetime
    interval: int

    def serialize(self) -> _JSONDict:
        """Serialize the dataclass to a JSON-compatible dictionary."""
        # Create basic serialization dictionary
        d = super().serialize()
        # Replace datetime with ISO format string
        d["expires_at"] = d["expires_at"].isoformat()
        return d

    @classmethod
    def deserialize(cls, data: _JSONDict) -> _PodDeviceCodeDetails:
        """Deserialize the dataclass from a JSON-compatible dictionary."""
        # Reconstruct datetime object
        data["expires_at"] = datetime.fromisoformat(data["expires_at"])
        return cls(**data)


@dataclass
class _ModellerDeviceCodeDetails:
    """The modeller-related device code response details."""

    user_code: str
    verification_uri: str
    verification_uri_complete: str


class CommunicationDetails(NamedTuple):
    """Container for received communication details."""

    mailbox_id: str
    pod_mailbox_ids: Dict[str, str]
    task_id: str



"""Workers for handling task running on pods."""
from __future__ import annotations

from typing import Any, Optional, cast

import sqlvalidator

from bitfount.data.datasources.base_source import BaseSource
from bitfount.data.datasources.database_source import DatabaseSource
from bitfount.data.datastructure import DataStructure
from bitfount.data.exceptions import DataStructureError
from bitfount.federated.algorithms.model_algorithms.base import (
    _BaseModelAlgorithm,
    _BaseModelAlgorithmFactory,
)
from bitfount.federated.authorisation_checkers import _AuthorisationChecker
from bitfount.federated.logging import _get_federated_logger
from bitfount.federated.monitoring.monitor import task_config_update
from bitfount.federated.pod_vitals import _PodVitals
from bitfount.federated.privacy.differential import DPPodConfig
from bitfount.federated.protocols.base import _BaseProtocolFactory
from bitfount.federated.transport.message_service import _BitfountMessageType
from bitfount.federated.transport.worker_transport import _WorkerMailbox
from bitfount.federated.types import SerializedProtocol, _DataLessAlgorithm
from bitfount.federated.utils import _PROTOCOLS
from bitfount.hub.api import BitfountHub
from bitfount.schemas.utils import bf_load
from bitfount.types import _JSONDict

logger = _get_federated_logger(__name__)


class _Worker:
    """Client worker which runs a protocol locally.

    Args:
        datasource: BaseSource object.
        mailbox: Relevant mailbox.
        bitfounthub: BitfountHub object.
        authorisation: AuthorisationChecker object.
        pod_identifier: Identifier of the pod the Worker is running in.
        serialized_protocol: SerializedProtocol dictionary that the Pod has received
            from the Modeller.
        pod_vitals: PodVitals object.
        pod_dp: DPPodConfig object.
    """

    def __init__(
        self,
        datasource: BaseSource,
        mailbox: _WorkerMailbox,
        bitfounthub: BitfountHub,
        authorisation: _AuthorisationChecker,
        pod_identifier: str,
        serialized_protocol: SerializedProtocol,
        pod_vitals: Optional[_PodVitals] = None,
        pod_dp: Optional[DPPodConfig] = None,
        **_kwargs: Any,
    ):
        self.datasource = datasource
        self.mailbox = mailbox
        self.hub = bitfounthub
        self.authorisation: _AuthorisationChecker = authorisation
        self.pod_identifier = pod_identifier
        self.serialized_protocol = serialized_protocol
        self.pod_vitals = pod_vitals
        self._pod_dp = pod_dp

    async def run(self) -> None:
        """Calls relevant training procedure and sends back weights/results."""
        # Send task to Monitor service. This is done regardless of whether or not
        # the task is accepted. This method is being run in a task monitor context
        # manager so no need to set the task monitor prior to sending.
        task_config_update(dict(self.serialized_protocol))

        # Check authorisation with access manager
        authorisation_errors = await self.authorisation.check_authorisation()

        if authorisation_errors.messages:
            # Reject task, as there were errors
            await self.mailbox.reject_task(
                authorisation_errors.messages,
            )
            return

        # Accept task and inform modeller
        logger.info("Task accepted, informing modeller.")
        await self.mailbox.accept_task()

        # Wait for Modeller to give the green light to start the task
        await self.mailbox.get_task_start_update()

        # Update hub instance if BitfountModelReference
        if model := self.serialized_protocol["algorithm"].get("model"):
            if model["class_name"] == "BitfountModelReference":
                logger.debug("Patching model reference hub.")
                model["hub"] = self.hub

        # Deserialize protocol only after task has been accepted just to be safe
        protocol = cast(
            _BaseProtocolFactory,
            bf_load(cast(_JSONDict, self.serialized_protocol), _PROTOCOLS),
        )

        # Load data for training here
        datastructure: Optional[DataStructure] = None

        if isinstance(protocol.algorithm, _BaseModelAlgorithmFactory):
            datastructure = protocol.algorithm.model.datastructure

        if not isinstance(protocol.algorithm, _DataLessAlgorithm):
            # We execute the query directly on the db connection,
            # or load the data at runtime for a csv.
            # TODO: [NO_TICKET: Reason] No ticket created yet. Add the private sql query algorithm here as well. # noqa: B950
            self._load_data_for_worker(datastructure=datastructure)

        # Calling the `worker` method on the protocol also calls the `worker` method on
        # underlying objects such as the algorithm and aggregator. The algorithm
        # `worker` method will also download the model from the Hub if it is a
        # `BitfountModelReference`
        worker_protocol = protocol.worker(mailbox=self.mailbox, hub=self.hub)

        # If the algorithm is a model algorithm, then we need to pass the pod identifier
        # to the model so that it can extract the relevant information from the
        # datastructure the Modeller has sent.
        if isinstance(worker_protocol.algorithm, _BaseModelAlgorithm):
            worker_protocol.algorithm.model.set_pod_identifier(self.pod_identifier)

        await worker_protocol.run(
            datasource=self.datasource,
            pod_dp=self._pod_dp,
            pod_vitals=self.pod_vitals,
            pod_identifier=self.mailbox.pod_identifier,
        )
        logger.info("Task complete.")
        self.mailbox.delete_all_handlers(_BitfountMessageType.LOG_MESSAGE)

    def _load_data_for_worker(
        self,
        datastructure: Optional[DataStructure] = None,
    ) -> None:
        """Load the data for the worker."""
        sql_query: Optional[str] = None
        table: Optional[str] = None
        kwargs = {}

        if datastructure:
            if datastructure.table:
                if isinstance(datastructure.table, dict):
                    if not (table := datastructure.table.get(self.pod_identifier)):
                        raise DataStructureError(
                            f"Table definition not found for {self.pod_identifier}. "
                            f"Table definitions provided in this DataStructure: "
                            f"{str(datastructure.table)}"
                        )
                    kwargs["table_name"] = table
                elif isinstance(datastructure.table, str):
                    kwargs["table_name"] = datastructure.table
            elif datastructure.query:
                if isinstance(datastructure.query, dict):
                    if not (sql_query := datastructure.query.get(self.pod_identifier)):
                        raise DataStructureError(
                            f"Query definition not found for {self.pod_identifier}. "
                            f"Query definitions provided in this DataStructure: "
                            f"{str(datastructure.query)}"
                        )
                elif isinstance(datastructure.query, str):
                    sql_query = datastructure.query
                if sql_query and sqlvalidator.parse(sql_query).is_valid():
                    if not isinstance(self.datasource, DatabaseSource):
                        raise ValueError(
                            "Incompatible DataStructure, data source pair. "
                            "DataStructure is expecting the data source to "
                            "be a DatabaseSource."
                        )
                    kwargs["sql_query"] = sql_query
        # This call loads the data for a multi-table BaseSource as specified by the
        # Modeller/DataStructure.
        self.datasource.load_data(**kwargs)



"""Handles messages for a Pods and Modellers at the protocol level during a task."""
import asyncio
import logging
from typing import (
    Any,
    Callable,
    Dict,
    Final,
    Iterable,
    List,
    Mapping,
    Optional,
    Tuple,
    Union,
)
import warnings

from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey, RSAPublicKey
import msgpack

from bitfount import config
from bitfount.federated.encryption import _RSAEncryption
from bitfount.federated.logging import _federate_logger, _get_federated_logger
from bitfount.federated.transport.base_transport import (
    Handler,
    _BaseMailbox,
    _send_aes_encrypted_message,
)
from bitfount.federated.transport.handlers import (
    _AsyncMultipleResponsesHandler,
    _get_message_awaitable,
    _OnlineResponseHandling,
)
from bitfount.federated.transport.message_service import (
    _BitfountMessage,
    _BitfountMessageType,
    _DecryptedBitfountMessage,
    _MessageService,
)
from bitfount.federated.transport.types import (
    _OIDCAuthFlowResponse,
    _OIDCClientID,
    _PodDeviceCodeDetails,
)
from bitfount.federated.types import _PodResponseType
from bitfount.types import _JSONDict, _SAMLResponse, _SerializedWeights

logger = _get_federated_logger(__name__)

# How long to wait for Modeller authentication response
_DEFAULT_AUTHENTICATION_MODELLER_RESPONSE_TIMEOUT: Final[int] = 5 * 60  # 5 minutes
# How long in seconds to wait for a response from the modeller before checking if
# the modeller is still online.
_SOFT_LIMIT_MESSAGE_TIMEOUT: Final[int] = config.BITFOUNT_ONLINE_CHECK_SOFT_LIMIT
# How long in seconds to wait for a response from the modeller regarding their
# online status before aborting the task. The hard limit can only be reached after
# the soft limit has already been reached.
_HARD_LIMIT_MESSAGE_TIMEOUT: Final[int] = config.BITFOUNT_ONLINE_CHECK_HARD_LIMIT


class _WorkerMailbox(_BaseMailbox):
    """Used by a pod for handling messages during a task."""

    def __init__(
        self,
        pod_identifier: str,
        modeller_mailbox_id: str,
        modeller_name: str,
        aes_encryption_key: bytes,
        message_service: _MessageService,
        pod_mailbox_ids: Mapping[str, str],
        task_id: Optional[str] = None,
        handlers: Optional[
            Mapping[_BitfountMessageType, Union[Handler, Iterable[Handler]]]
        ] = None,
    ):
        """Create new worker mailbox for a specific task.

        Args:
            pod_identifier: identifier for the pod that contains this worker.
            modeller_mailbox_id: mailbox id for modeller involved in the task.
            modeller_name: name of the modeller involved in the task.
            aes_encryption_key: encryption key for task messages.
            message_service: the underlying message service.
            pod_mailbox_ids: mapping of pod_identifier to worker mailbox IDs for
                             all pods involved in the task.
            task_id: The ID of the task that this mailbox is associated with.
            handlers: an optional mapping of message types to handlers to initialise
                      with.
        """
        # Our own mailbox ID is stored in the pods mailbox dict.
        super().__init__(
            mailbox_id=pod_mailbox_ids[pod_identifier],
            message_service=message_service,
            handlers=handlers,
        )
        self.pod_identifier = pod_identifier
        self.modeller_mailbox_id = modeller_mailbox_id
        self.modeller_name = modeller_name
        self.aes_encryption_key = aes_encryption_key
        self.pod_mailbox_ids: Dict[str, str] = dict(pod_mailbox_ids)
        self._task_id = task_id
        if self._task_id is None:
            warnings.warn(
                "Worker mailboxes should be instantiated with task IDs;"
                " ask the Modeller to upgrade their bitfount version if possible.",
                FutureWarning,
            )
            logger.warning(
                "Worker mailboxes should be instantiated with task IDs;"
                " ask the Modeller to upgrade their bitfount version if possible."
            )

        # Create modeller online checker and set it to monitor all messages
        self._online_response_handler = _OnlineResponseHandling(
            self.modeller_name, self.aes_encryption_key
        )
        # TODO: [BIT-2248] Only need to register this as universal once dispatch
        #       is more permissible.
        self.register_universal_handler(
            self._online_response_handler.response_handler, high_priority=True
        )
        self.register_handler(
            _BitfountMessageType.ONLINE_RESPONSE,
            self._online_response_handler.response_handler,
            high_priority=True,
        )

        # Gather list of other pods. Iteration order is important here (to ensure
        # consistency between share generation order), so we use List rather than
        # another collection.
        self.other_pods: List[str] = [
            pod_identifier
            for pod_identifier in pod_mailbox_ids
            if pod_identifier != self.pod_identifier
        ]
        self._setup_federated_logging()

    @property
    def task_id(self) -> Optional[str]:
        """The task ID of the task associated with this mailbox."""
        return self._task_id

    async def accept_task(self) -> None:
        """Sends an acceptance message to the modeller."""
        await self._send_aes_encrypted_message(
            {_PodResponseType.ACCEPT.name: self.pod_identifier},
            _BitfountMessageType.JOB_ACCEPT,
        )

    async def reject_task(
        self,
        error_messages: Mapping[str, Iterable[str]],
    ) -> None:
        """Send a rejection of a training request to a modeller.

        Args:
            error_messages: Error messages to send

        Returns:
            True if message sent successfully, else false

        """
        logger.info(f"Rejecting task from {self.modeller_name}")
        logger.debug(
            f"Rejected task from {self.modeller_name} "
            f"at mailbox: {self.modeller_mailbox_id}"
        )
        await self._send_aes_encrypted_message(
            error_messages, _BitfountMessageType.JOB_REJECT
        )

    async def issue_saml_challenge(
        self,
        saml_request: str,
    ) -> None:
        """Send SAML challenge to a modeller."""
        logger.info(f"Issuing SAML Challenge to: '{self.modeller_name}'.")
        await self._send_aes_encrypted_message(
            saml_request, _BitfountMessageType.SAML_REQUEST
        )

    async def get_saml_response(
        self, timeout: Optional[int] = _DEFAULT_AUTHENTICATION_MODELLER_RESPONSE_TIMEOUT
    ) -> _SAMLResponse:
        """Awaits for the SAML Response from the modeller."""
        saml_response: _SAMLResponse = await self._get_message_and_decrypt(
            _BitfountMessageType.SAML_RESPONSE, timeout
        )
        return saml_response

    async def send_oidc_client_id(self, client_id: str) -> None:
        """Sends the Client ID to use for OIDC authentication to modeller."""
        logger.info("Sending client ID to modeller for OIDC authentication.")
        await self._send_aes_encrypted_message(
            _OIDCClientID(client_id).serialize(),
            _BitfountMessageType.OIDC_CHALLENGE,
        )

    async def get_oidc_auth_flow_response(
        self, timeout: Optional[int] = _DEFAULT_AUTHENTICATION_MODELLER_RESPONSE_TIMEOUT
    ) -> _OIDCAuthFlowResponse:
        """Get OIDC Auth Code Flow response from the modeller.

        Response will contain an authorization code, code verifier, and redirect URI.

        Args:
            timeout: How long to wait for response, in seconds. If `None`, will wait
                indefinitely. Defaults to 5 minutes.

        Returns:
            Tuple of authorization code, code verifier, and redirect URI.
        """
        logger.info(f"Awaiting OIDC response from '{self.modeller_name}'")
        decrypted: _JSONDict = await self._get_message_and_decrypt(
            _BitfountMessageType.OIDC_AFC_PKCE_RESPONSE, timeout
        )

        # Check contents
        try:
            return _OIDCAuthFlowResponse.deserialize(decrypted)
        except (KeyError, TypeError):
            try:
                # Assume it's a dict but that we just couldn't deserialize it
                raise KeyError(
                    f"Expected auth_code, code_verifier, and redirect_uri to be in "
                    f"OIDC response; got {decrypted.keys()}"
                )
            except AttributeError:
                # If not, raise a TypeError
                raise TypeError(
                    f"Unable to access OIDC response contents; expected dict, "
                    f"got {type(decrypted)}"
                )

    async def get_oidc_device_code_response(
        self, timeout: Optional[int] = _DEFAULT_AUTHENTICATION_MODELLER_RESPONSE_TIMEOUT
    ) -> _PodDeviceCodeDetails:
        """Get OIDC Device Code Flow response from the modeller.

        Response will contain a device code, and polling details.

        Args:
            timeout: How long to wait for response, in seconds. If `None`, will wait
                indefinitely. Defaults to 5 minutes.

        Returns:
            Tuple device code, when the code expires, and the polling interval to use.
        """
        logger.info(f"Awaiting OIDC device response from '{self.modeller_name}'")
        decrypted: _JSONDict = await self._get_message_and_decrypt(
            _BitfountMessageType.OIDC_DEVICE_CODE_RESPONSE, timeout
        )

        # Check contents and decode as needed
        try:
            return _PodDeviceCodeDetails.deserialize(decrypted)
        except (KeyError, TypeError):
            try:
                # Assume it's a dict but that we just couldn't deserialize it
                raise KeyError(
                    f"Expected device_code, expires_at, and interval to be in "
                    f"OIDC response; got {decrypted.keys()}"
                )
            except AttributeError:
                # If not, raise a TypeError
                raise TypeError(
                    f"Unable to access OIDC response contents; expected dict, "
                    f"got {type(decrypted)}"
                )

    async def _send_aes_encrypted_message(
        self, object_to_send: Any, message_type: _BitfountMessageType
    ) -> None:
        """Send message to modeller, AES encrypted.

        Args:
            object_to_send: Body of the message (not encrypted)
            message_type: The type of the message to send
        """
        await _send_aes_encrypted_message(
            object_to_send,
            self.aes_encryption_key,
            self.message_service,
            message_type=message_type,
            recipient=self.modeller_name,
            recipient_mailbox_id=self.modeller_mailbox_id,
            sender=self.pod_identifier,
            sender_mailbox_id=self.mailbox_id,
            task_id=self._task_id,
        )

    async def send_evaluation_results(self, eval_results: Mapping[str, float]) -> None:
        """Sends evaluation results to the modeller."""
        logger.info("Sending results to modeller...")
        await self._send_aes_encrypted_message(
            eval_results, _BitfountMessageType.EVALUATION_RESULTS
        )

    async def _get_message(
        self, message_type: _BitfountMessageType, timeout: Optional[int]
    ) -> _BitfountMessage:
        """Generic handler for single message retrieval.

        Args:
            message_type: The type of message to wait on.
            timeout: How long to wait before timing out.

        Returns:
            The message.

        Raises:
            asyncio.TimeoutError: If the message is not received within the timeout.
        """
        logger.debug(
            f"Waiting for message ({message_type}) retrieval in"
            f" {self.__class__.__name__}._get_message()"
        )

        # Registers the handler for the expected message type
        async_awaitable = _get_message_awaitable()
        self.register_temp_handler(message_type, async_awaitable)

        # Waits for the message until a `TimeoutError` occurs. If a message is received
        # before the timeout, it is simply returned. Otherwise, a message is sent to the
        # modeller to check that they are still online. If they are online, the timeout
        # resets and we wait for the original message again. If they are not online,
        # this will raise another `TimeoutError` waiting for the Modeller's response. On
        # the second timeout, the task is cancelled by sending a `TASK_ABORT` message to
        # the modeller and re-raising the original `TimeoutError`.
        while True:
            try:
                message = await async_awaitable.result(timeout)
                break
            except asyncio.TimeoutError as te:
                logger.debug(
                    f"Soft limit timeout hit waiting for message {message_type.name}."
                )

                logger.info("Checking if Modeller is still online...")
                online_check_uuid = self._online_response_handler.get_online_check_id()
                await self.check_modeller_online(online_check_uuid)

                # Want to wait for _either_ the original message to come through or
                # for the modeller to indicate it is online.
                message_result_task = asyncio.create_task(async_awaitable.result())
                online_response_task = asyncio.create_task(
                    self._online_response_handler.wait_for_response(online_check_uuid)
                )
                joint_wait = asyncio.wait(
                    (message_result_task, online_response_task),
                    return_when=asyncio.FIRST_COMPLETED,
                )

                try:
                    done, pending = await asyncio.wait_for(
                        joint_wait, timeout=_HARD_LIMIT_MESSAGE_TIMEOUT
                    )
                except asyncio.TimeoutError:
                    logger.warning("Modeller is offline. Aborting task.")
                    await self.send_task_abort_message()

                    # Double-check these tasks are cancelled
                    message_result_task.cancel()
                    online_response_task.cancel()

                    raise te
                else:
                    if message_result_task in done:
                        # We have our message, cancel the response waiter and return
                        logger.info(
                            "Modeller is online, responded with expected message."
                        )
                        online_response_task.cancel()
                        self._online_response_handler.remove_waiter(online_check_uuid)
                        return message_result_task.result()
                    else:  # online_response_task in done
                        logger.info("Modeller is online, continuing to wait.")

                        # One final check to see if it's there
                        if not message_result_task.done():
                            # Cancel message result task as we're going to start the
                            # loop again and await on it there
                            message_result_task.cancel()
                        else:
                            return message_result_task.result()

        return message

    def _aes_decrypt(self, message: _BitfountMessage) -> Any:
        """Decrypt message from modeller.

        Args:
            message: Encrypted message to decrypt.

        Returns:
            Decrypted message body.
        """
        return message.decrypt(self.aes_encryption_key).body

    async def _get_message_and_decrypt(
        self,
        message_type: _BitfountMessageType,
        timeout: Optional[int] = _SOFT_LIMIT_MESSAGE_TIMEOUT,
    ) -> Any:
        """Generic handler for single encrypted message retrieval.

        Args:
            message_type: The type of message to wait on.
            timeout: How long to wait before timing out.

        Returns:
            The decrypted message contents.
        """
        message = await self._get_message(message_type, timeout)
        return self._aes_decrypt(message)

    async def get_training_iteration_complete_update(
        self, timeout: Optional[int] = _SOFT_LIMIT_MESSAGE_TIMEOUT
    ) -> bool:
        """Awaits for the next training complete update from the modeller."""
        training_complete: bool = await self._get_message_and_decrypt(
            _BitfountMessageType.TRAINING_COMPLETE, timeout=timeout
        )
        return training_complete

    async def get_task_start_update(
        self, timeout: Optional[int] = _SOFT_LIMIT_MESSAGE_TIMEOUT
    ) -> None:
        """Awaits for the task start message from the modeller."""
        await self._get_message_and_decrypt(
            _BitfountMessageType.TASK_START, timeout=timeout
        )

    async def get_task_complete_update(
        self, timeout: Optional[int] = _SOFT_LIMIT_MESSAGE_TIMEOUT
    ) -> None:
        """Awaits for the task complete message from the modeller."""
        await self._get_message_and_decrypt(
            _BitfountMessageType.TASK_COMPLETE, timeout=timeout
        )

    async def check_modeller_online(self, online_check_uuid: str) -> None:
        """Send message to Modeller to check if the Modeller is online."""
        await self._send_aes_encrypted_message(
            online_check_uuid, _BitfountMessageType.ONLINE_CHECK
        )

    async def send_task_abort_message(self) -> None:
        """Send task abort message to Modeller."""
        await self._send_aes_encrypted_message(None, _BitfountMessageType.TASK_ABORT)

    async def log(self, message: Mapping[str, object]) -> None:
        """Log message to Modeller."""
        await self._send_aes_encrypted_message(
            message, _BitfountMessageType.LOG_MESSAGE
        )

    def _setup_federated_logging(self) -> None:
        """Set up federated logging."""
        _federate_logger(self)

        def log_message_handler(message: _BitfountMessage) -> None:
            log_message: _JSONDict = self._aes_decrypt(message)
            # We prepend the log message to show it's come from the Modeller
            log_message["msg"] = f"Modeller: {log_message['msg']}"
            # We remove the `federated` key to avoid recursively sending a federated
            # log message on both the Modeller and Worker sides
            log_message.pop("federated")
            logger.handle(logging.makeLogRecord(log_message))

        self.register_handler(_BitfountMessageType.LOG_MESSAGE, log_message_handler)


async def _send_training_metrics(
    validation_metrics: Mapping[str, str],
    worker_mailbox: _WorkerMailbox,
) -> None:
    """Sends a model parameter update to the modeller.

    Args:
        validation_metrics: The relevant metrics at training iteration.
        worker_mailbox: The worker mailbox to use to send the parameter update.
    """
    # Deliberate access to private method here as that method shouldn't be used in
    # any other context than transport layer access.
    # noinspection PyProtectedMember
    await worker_mailbox._send_aes_encrypted_message(
        validation_metrics, _BitfountMessageType.TRAINING_METRICS
    )


async def _send_parameter_update(
    parameter_update: _SerializedWeights, worker_mailbox: _WorkerMailbox
) -> None:
    """Sends a model parameter update to the modeller.

    Args:
        parameter_update: The parameter update to send, already serialized.
        worker_mailbox: The worker mailbox to use to send the parameter update.
    """
    # Deliberate access to private method here as that method shouldn't be used in
    # any other context than transport layer access.
    # noinspection PyProtectedMember
    await worker_mailbox._send_aes_encrypted_message(
        parameter_update, _BitfountMessageType.TRAINING_UPDATE
    )


async def _send_public_key(
    public_key: RSAPublicKey, worker_mailbox: _WorkerMailbox
) -> None:
    """Sends the public key to the modeller.

    Used for the PSI protocol.

    Args:
        public_key: The public key to use in the PSI.
        worker_mailbox: The worker mailbox to use to send the parameter update.
    """
    # Serialize the public key first
    serialized_key = public_key.public_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PublicFormat.SubjectPublicKeyInfo,
    )
    await worker_mailbox._send_aes_encrypted_message(
        serialized_key, _BitfountMessageType.KEY_EXCHANGE
    )


async def _send_psi_dataset_worker(
    dataset: Tuple[List[str], List[str]], worker_mailbox: _WorkerMailbox
) -> None:
    """Sends psi datasets to the modeller."""
    await worker_mailbox._send_aes_encrypted_message(
        dataset, _BitfountMessageType.PSI_DATASET
    )


async def _get_model_parameters(
    worker_mailbox: _WorkerMailbox, timeout: Optional[int] = _SOFT_LIMIT_MESSAGE_TIMEOUT
) -> _SerializedWeights:
    """Awaits for the next model parameter update from the modeller.

    Args:
        worker_mailbox: The worker mailbox that the message will be sent to.
        timeout: Optional. The time to wait in seconds for the next model
            parameter update.

    Returns:
        The received parameter update.
    """
    # Await messages for updates sent modeller->worker
    # Deliberate access to private method here as that method shouldn't be used in
    # any other context than transport layer access.
    # noinspection PyProtectedMember
    weights: _SerializedWeights = await worker_mailbox._get_message_and_decrypt(
        _BitfountMessageType.MODEL_PARAMETERS, timeout=timeout
    )

    return weights


async def _get_psi_dataset(
    worker_mailbox: _WorkerMailbox, timeout: Optional[int] = _SOFT_LIMIT_MESSAGE_TIMEOUT
) -> List[str]:
    """Awaits for the psi dataset from the modeller.

    Args:
        worker_mailbox: The worker mailbox that the message will be sent to.
        timeout: Optional. The time to wait in seconds for the next model
            parameter update.

    Returns:
        The psi dataset.
    """
    psi_dataset: List[str] = await worker_mailbox._get_message_and_decrypt(
        _BitfountMessageType.PSI_DATASET, timeout=timeout
    )
    return psi_dataset


class _InterPodWorkerMailbox(_WorkerMailbox):
    """A worker mailbox that also handles pod-to-pod communication.

    In particular ensures that other pod RSA keys are stored and used.
    """

    def __init__(
        self,
        pod_public_keys: Mapping[str, RSAPublicKey],
        private_key: RSAPrivateKey,
        *args: Any,
        **kwargs: Any,
    ):
        """Create new inter-pod worker mailbox for a specific task.

        Args:
            pod_public_keys: Mapping of pod_identifier to the pod's RSA public
                             key for all pods involved in the task. This is for
                             inter-pod communication.
            private_key: The RSA private key for this pod.
            *args: Positional arguments as per _WorkerMailbox.
            **kwargs: Keyword arguments as per _WorkerMailbox.
        """
        super().__init__(*args, **kwargs)
        self._pod_public_keys = pod_public_keys

        # Check that we have public keys for all other pods
        missing_keys = set(self.other_pods) - set(self._pod_public_keys.keys())
        if missing_keys:
            missing_keys_str = ", ".join(missing_keys)
            raise ValueError(
                f"We are missing public keys for the following pods: "
                f"{missing_keys_str}. "
                f"Unable to continue inter-pod communication."
            )

        self._private_key = private_key

    async def _send_pod_to_pod_message(
        self,
        recipient: str,
        recipient_mailbox_id: str,
        object_to_send: Any,
        message_type: _BitfountMessageType,
    ) -> None:
        """Send message to other pod/worker, unencrypted.

        Args:
            recipient: The identifier of the worker/pod.
            recipient_mailbox_id: The mailbox to send message to.
            object_to_send: Body of the message (not encrypted).
            message_type: The type of the message to send.
        """
        try:
            recipient_key = self._pod_public_keys[recipient]
        except KeyError:
            logging.error(
                f"Unable to find public key for pod {recipient}. "
                f"Unable to send pod-to-pod message."
            )
            return None

        message_body: bytes = _RSAEncryption.encrypt(
            msgpack.dumps(object_to_send), recipient_key
        )
        await self.message_service.send_message(
            _BitfountMessage(
                message_type=message_type,
                body=message_body,
                recipient=recipient,
                recipient_mailbox_id=recipient_mailbox_id,
                sender=self.pod_identifier,
                sender_mailbox_id=self.mailbox_id,
                task_id=self._task_id,
            ),
            already_packed=True,
        )

    def _pod_to_pod_message_handler(
        self, message: _BitfountMessage
    ) -> _DecryptedBitfountMessage:
        """Handler for decrypting pod-to-pod messages."""
        return message.decrypt_rsa(self._private_key)


async def _send_secure_shares_to_others(
    secure_share_generator: Callable[[], int],
    worker_mailbox: _InterPodWorkerMailbox,
) -> None:
    """Sends result of `secure_share_generator` to all other pods in training task.

    Args:
        secure_share_generator: The function to be called which returns
                                the secure share to be sent.
        worker_mailbox: The worker mailbox to send the shares with.
    """
    for worker_identifier in worker_mailbox.other_pods:
        worker_mailbox_id = worker_mailbox.pod_mailbox_ids[worker_identifier]

        logger.debug(
            f"Sending secure share to mailbox of other worker "
            f"(from {worker_mailbox.pod_identifier} to {worker_identifier}): "
            f"{worker_mailbox_id}"
        )

        share = secure_share_generator()
        # Deliberate access to private method here as that method shouldn't be used in
        # any other context than transport layer access.
        # noinspection PyProtectedMember
        await worker_mailbox._send_pod_to_pod_message(
            recipient=worker_identifier,
            recipient_mailbox_id=worker_mailbox_id,
            object_to_send=share,
            message_type=_BitfountMessageType.SECURE_SHARE,
        )


async def _get_worker_secure_shares(
    worker_mailbox: _InterPodWorkerMailbox, timeout: Optional[int] = None
) -> List[int]:
    """Awaits the set of secure shares from the other workers.

    Args:
        worker_mailbox: The worker mailbox the shares will be received at.
        timeout: Optional. The number of seconds to wait for secure shares to arrive.

    Returns:
        The list of received shares.
    """
    shares: List[int] = []

    # Create light-weight handler to append to shared list.
    # Note that the secure shares are NOT received as encrypted messages.
    def worker_secure_share_handler(message: _BitfountMessage) -> None:
        logger.debug(f"Receiving secure share from worker {message.sender}")
        # Deliberate access to private method here as that method shouldn't be used in
        # any other context than transport layer access.
        # noinspection PyProtectedMember
        decrypted_message = worker_mailbox._pod_to_pod_message_handler(message)
        share: int = decrypted_message.body
        shares.append(share)

    # Await on all the other workers to send their shares, which will be
    # appended to the list above.
    with _AsyncMultipleResponsesHandler(
        handler=worker_secure_share_handler,
        message_types=_BitfountMessageType.SECURE_SHARE,
        mailbox=worker_mailbox,
        responders=worker_mailbox.other_pods,
    ) as response_handler:
        await response_handler.wait_for_responses(timeout=timeout)

    return shares



"""Various identity verification related methods and classes."""
from typing import Final

# We run a temporary HTTP listener on this port
# Why 29206? It's unlikely to be in use.
# And it's 'Bitf' (2, 9, 20, 6)
_BITFOUNT_MODELLER_PORT: Final = 29206
_PORT_WAIT_TIMEOUT: Final = 10



"""Copyright 2019 Less Wright.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Mish: A Self Regularized Non-Monotonic Neural Activation Function
https://arxiv.org/abs/1908.08681v1
github: https://github.com/lessw2020/mish
"""

import torch
import torch.nn as nn
import torch.nn.functional as F


class Mish(nn.Module):
    """Mish: A Self Regularized Non-Monotonic Neural Activation Function.

    More information here:
        Paper: https://arxiv.org/abs/1908.08681
        Github: https://github.com/lessw2020/mish
    """

    def __init__(self) -> None:
        super().__init__()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward method.

        Args:
            x: Input to the activation layer.

        Returns:
            Output of the activation layer.
        """
        # inlining this saves 1 second per epoch (V100 GPU) vs having a temp x
        # and then returning x
        return x * (torch.tanh(F.softplus(x)))



#!/usr/bin/env python
"""Evaluate a pretrained local model on local data."""
import logging
from typing import cast

import fire
import numpy as np
import pandas as pd
import yaml

from bitfount import config
from bitfount.data.datasources.dataframe_source import DataFrameSource
from bitfount.data.datastructure import DataStructure
from bitfount.data.schema import BitfountSchema
from bitfount.federated.utils import _MODELS
from bitfount.metrics import MetricCollection
from bitfount.runners.utils import setup_loggers
from bitfount.transformations.dataset_operations import (
    CleanDataTransformation,
    NormalizeDataTransformation,
)
from bitfount.transformations.processor import TransformationProcessor

loggers = setup_loggers([logging.getLogger("bitfount")])

config._BITFOUNT_CLI_MODE = True


def evaluate_model(
    path_to_config_yaml: str, path_to_test_csv: str, path_to_model: str
) -> None:
    """Evaluates a model's performance."""
    with open(path_to_config_yaml) as f:
        config = yaml.safe_load(f)
    algorithm = config.pop("algorithm")
    schema_filename = config.pop("schema")
    batch_size = config.get("batch_size")
    datasource_args = config["datasource_args"]
    target = config["target"]

    # Load dataset
    data = pd.read_csv(path_to_test_csv)
    schema = BitfountSchema.load_from_file(schema_filename)
    if len(schema.tables) > 1:
        raise ValueError("Only single-table schemas are supported.")

    datasource = DataFrameSource(data=data, **datasource_args)
    # Transform dataset
    clean_data = CleanDataTransformation()
    normalize = NormalizeDataTransformation()
    processor = TransformationProcessor([clean_data, normalize], schema.tables[0])
    datasource.data = processor.transform(datasource.data)

    # Create datastructure and get test dataloader
    data_structure = DataStructure(target=target, table=schema.table_names[0])
    # TODO: [BIT-1167] process transformations here

    # Load model
    neural_network = _MODELS[algorithm](datastructure=data_structure, schema=schema)
    neural_network.initialise_model(data=datasource)
    test_data_loader = neural_network.databunch.get_train_dataloader(
        batch_size=batch_size
    )

    neural_network.deserialize(path_to_model)

    # Evaluate model on test dataloader
    preds, targs = neural_network.evaluate(test_data_loader)
    # TODO: [BIT-1604] Remove these cast statements once they become superfluous.
    preds = cast(np.ndarray, preds)
    targs = cast(np.ndarray, targs)
    metrics = MetricCollection.create_from_model(neural_network)
    results = metrics.compute(targs, preds)
    logging.info(str(results))


def main() -> None:
    """Script entry point."""
    fire.Fire(evaluate_model)


if __name__ == "__main__":
    main()



"""Types related to monitoring modules."""
from enum import Enum
from typing import Optional, Protocol, TypedDict, runtime_checkable

from typing_extensions import NotRequired


@runtime_checkable
class HasTaskID(Protocol):
    """Protocol for describing objects that contain task IDs."""

    _task_id: Optional[str]


class MonitorRecordPrivacy(Enum):
    """Privacy options for monitor record.

    Note that these are primarily interpreted from the point of view of the
    pod owner.
    """

    PRIVATE = "PRIVATE"  # only the pod owner
    OWNER = PRIVATE  # alias to PRIVATE
    OWNER_MODELLER = "MODELLER"  # pod owner plus the modeller
    ALL_PARTICIPANTS = "ALL_PARTICIPANTS"  # everyone in the task


class AdditionalMonitorMessageTypes(Enum):
    """Additional monitoring message types.

    These are explicit message types in addition to the full set of
    _BitfountMessageType types being supported.
    """

    TASK_CONFIG = "TASK_CONFIG"
    TASK_STATUS_UPDATE = "TASK_STATUS_UPDATE"


class ProgressCounterDict(TypedDict):
    """Form of the progress counter dictionaries."""

    value: float
    total: NotRequired[float]
    estimated_total: NotRequired[float]
    unit: NotRequired[str]
